/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "schema.h"

long debug_helper_mem_allocated;
int max_xpath_seq, *xpath_prefix;
int p_curr_xpath;

/* function templates */
void helper_calc_tree_paths();
int  count_helper_xpaths(int depth, int i);
void calc_max_xpath_seq(int i);
void populate_helper_xpaths(int i);
void add_helper_branch(int i, int h, int branch);
int  add_helper_elements(int i, int i_subr, int h);
int  multiply_helper_elements(int i, int i_subr, int h);
void calc_field_match_constraints();
void traverse_helper_paths(int i_helper, int h, int i, int flag_child_of_match_none);
void debug_print_helper_xpaths(char *msg, int i_helper, int h_helper);


void helper_calc_tree_paths()
{
  /* Find all the combinations of paths for each child of the parent element */
  int h, i, p;

  if (sw_debug)
  {
    debug_helper_mem_allocated = 0;
    fprintf(stderr, "DEBUG: Helpers: Calculating traversal paths and allocating memory.%s", newline);
  }

  max_xpath_seq = 0;

  /* allocate resources for the helpers */
  for (i = schema_match_start_level ; i < schema_parent.nelements ; i++)
  {
    if (schema_element[i].element_helper == NULL)
    {
      /* allocate element_helper memory */
      schema_element[i].element_helper = (struct_element_helper *) malloc(sizeof(struct_element_helper));
      if (schema_element[i].element_helper == NULL) out_of_mem();
      if (sw_debug)
        debug_helper_mem_allocated += sizeof(struct_element_helper);
      /* this gets re-adjusted by count_helper_xpaths() which needs all helpers allocated */
      schema_element[i].element_helper->max_xpath_seq = 1;
      schema_element[i].element_helper->complete = FALSE;
    }
  } /* end for/i */

  /* calculate size paramets and allocate memory */
  for (i = schema_match_start_level ; i < schema_parent.nelements ; i++)
  {
    /* calculate number of traversal paths of element tree */
    schema_element[i].element_helper->nxpaths = count_helper_xpaths(schema_match_start_level, i);
    /* calculate the size of the xpath array */
    calc_max_xpath_seq(i);
    /* allocate match_helper memory */
    schema_element[i].element_helper->match_helper = (struct_match_helper *) malloc(schema_element[i].element_helper->nxpaths * sizeof(struct_match_helper));
    if (schema_element[i].element_helper->match_helper == NULL) out_of_mem();
    if (sw_debug)
      debug_helper_mem_allocated += schema_element[i].element_helper->nxpaths * sizeof(struct_match_helper);

    /* allocate xpath memory of match_helper */
    for (h = 0 ; h < schema_element[i].element_helper->nxpaths ; h++)
    {
      schema_element[i].element_helper->match_helper[h].xpath = (int *) malloc(schema_element[i].element_helper->max_xpath_seq * sizeof(int));
      if (schema_element[i].element_helper->match_helper[h].xpath == NULL) out_of_mem();
      if (sw_debug)
        debug_helper_mem_allocated += schema_element[i].element_helper->max_xpath_seq * sizeof(int);
      schema_element[i].element_helper->match_helper[h].min  = 0;
      schema_element[i].element_helper->match_helper[h].max  = 0;
      schema_element[i].element_helper->match_helper[h].last = 0;
      /* init the xpath array */
      for (p = 0 ; p < schema_element[i].element_helper->max_xpath_seq ; p++)
        schema_element[i].element_helper->match_helper[h].xpath[p] = -1;
    } /* end for/h */
  } /* end for/i */

  /* reset again for further processing */
  for (i = schema_match_start_level ; i < schema_parent.nelements ; i++)
    schema_element[i].element_helper->complete = FALSE;

  /* allocate scratch mem for xpath processing */
  xpath_prefix = (int *) malloc(max_xpath_seq * sizeof(int));
  if (xpath_prefix == NULL) out_of_mem();
  /* populate the helpers */
  for (i = schema_match_start_level ; i < schema_parent.nelements ; i++)
    populate_helper_xpaths(i);
  (void) free(xpath_prefix);

  calc_field_match_constraints();

  if (sw_debug)
  {
    fprintf(stderr, "DEBUG: Helpers: Total helper memory allocated = %ld bytes.%s", debug_helper_mem_allocated, newline);
    debug_print_helper_xpaths("Match helper =", -1, -1);
  }
}


int count_helper_xpaths(depth, i)
int depth;
int i;
{
  /* Scan the element and its records looking for subrecords and recursively re-enter
   * while counting the number of unique paths taken until there are no more unique 
   * paths to be taken.
   * The resulting figure is used to determine the helper array size for each element */
  int i_subr, j, k;
  int count, col_count, row_count;
  int subrecord_count;

  col_count = schema_element[i].nrecords;
  row_count = 0;
  count     = 0;

  for (j = 0 ; j < schema_element[i].nrecords ; j++)
  {
    /* check if multiple subrecords in record */
    subrecord_count = 0;
    for (k = 0 ; k < schema_element[i].record[j]->nfields ; k++)
      if (schema_element[i].record[j]->field[k].type == FIELD_SUBRECORD)
        subrecord_count++;

    for (k = 0 ; k < schema_element[i].record[j]->nfields ; k++)
    {
      if (schema_element[i].record[j]->field[k].type == FIELD_SUBRECORD)
      {
        for (i_subr = schema_match_start_level ; i_subr < schema_parent.nelements ; i_subr++)
        {
          if (xstrcmp(schema_element[i].record[j]->field[k].name, schema_element[i_subr].name) == 0)
          {
            count = count_helper_xpaths(depth + 1, i_subr);
            if (subrecord_count < 2)
            {
              col_count += count;
            }
            else
            {
              if (row_count == 0)
                row_count = 1;
              if (count)
                row_count *= count;
            }
            break; /* break for/i_subr */
          }
        } /* end for/i_subr */
      }
    } /* end for/k */
    col_count += row_count;
    row_count = 0;
    /* Normally paths = (r1 - 1) + (sr1 - 1) + ... + 1 but if multiplication is 
     * involved (i.e. subrecord_count > 1) then don't add the number 1 */
    if (subrecord_count)
      col_count--;
  } /* end for/j */

  return(col_count);
}


void calc_max_xpath_seq(i)
int i;
{
  /* Scan the element and its records looking for subrecords and recursively re-enter
   * while measuring the size of the xpath array required.
   * The resulting figure is used to determine the xpath array size for each element */
  int i_subr, j, k;
  int p_max1, p_max2;

  if (!schema_element[i].element_helper->complete)
  {
    p_max2 = schema_element[i].element_helper->max_xpath_seq;
    for (j = 0 ; j < schema_element[i].nrecords ; j++)
    {
      p_max1 = schema_element[i].element_helper->max_xpath_seq;

      for (k = 0 ; k < schema_element[i].record[j]->nfields ; k++)
      {
        if (schema_element[i].record[j]->field[k].type == FIELD_SUBRECORD)
        {
          for (i_subr = schema_match_start_level ; i_subr < schema_parent.nelements ; i_subr++)
          {
            if (xstrcmp(schema_element[i].record[j]->field[k].name, schema_element[i_subr].name) == 0)
            {
              calc_max_xpath_seq(i_subr);
              /* re-calculate size of xpath array */
              p_max1 += schema_element[i_subr].element_helper->max_xpath_seq;
              break; /* break for/i_subr */
            }
          } /* end for/i_subr */
        }
      } /* end for/k */
      /* set the element max_xpath_seq */
      if (p_max1 > p_max2)
        p_max2 = p_max1;
    } /* end for/j */

    /* set the element max_xpath_seq */
    if (p_max2 > schema_element[i].element_helper->max_xpath_seq)
      schema_element[i].element_helper->max_xpath_seq = p_max2;

    /* set the global max_xpath_seq */
    if (schema_element[i].element_helper->max_xpath_seq > max_xpath_seq)
      max_xpath_seq = schema_element[i].element_helper->max_xpath_seq;

    /* as we bubble up set completion flag for this element i */
    schema_element[i].element_helper->complete = TRUE;
  }
}
  

void populate_helper_xpaths(i)
int i;
{
  /* Work to the bottom of the tree then build the xpath combinations
   * for each element while bubbling up the tree. */
  int i_subr, j, k, h, h_multiply_start;
  int flag_subrecord, subrecord_count;

  if (!schema_element[i].element_helper->complete)
  {
    h = 0;
    h_multiply_start = h;
    for (j = 0 ; j < schema_element[i].nrecords ; j++)
    {
      add_helper_branch(i, h, j);

      /* find out if any subrecords in this record */
      flag_subrecord = FALSE;
      for (k = 0 ; k < schema_element[i].record[j]->nfields ; k++)
      {
        if (schema_element[i].record[j]->field[k].type == FIELD_SUBRECORD)
        {
          flag_subrecord = TRUE;
          break;
        }
      }

      if (flag_subrecord)
      {
        /* process the fields of this subrecord */
        subrecord_count = 0;
        for (k = 0 ; k < schema_element[i].record[j]->nfields ; k++)
        {
          if (schema_element[i].record[j]->field[k].type == FIELD_SUBRECORD)
          {
            subrecord_count++;

            /* find the matching element */
            for (i_subr = schema_match_start_level ; i_subr < schema_parent.nelements ; i_subr++)
            {
              if (xstrcmp(schema_element[i].record[j]->field[k].name, schema_element[i_subr].name) == 0)
              {
                /* recursively drill down to the bottom of the tree before any arithmetic takes place */
                populate_helper_xpaths(i_subr);
                if (subrecord_count < 2) /* add subrecord xpaths */
                {
                  h_multiply_start = h;
                  h = add_helper_elements(i, i_subr, h);
                }
                else /* multiply subrecord paths */
                {
                  h = h_multiply_start;
                  h = multiply_helper_elements(i, i_subr, h);
                }
                break; /* break for/i_subr */
              }
            } /* end for/i_subr */
          }
        } /* end for/k */
      }
      else
      {
        if (h < schema_element[i].element_helper->nxpaths)
          h++;
      }
    } /* end for/j */
    /* as we bubble up set completion flag for this element i */
    schema_element[i].element_helper->complete = TRUE;
  }
}


void add_helper_branch(i, h, branch)
int i;
int h;
int branch;
{
  /* add the new branch to the helper array h of element i */
  int p;

  /* append branch (i.e. j (record)) to xpath array */
  for (p = 0 ; p < schema_element[i].element_helper->max_xpath_seq && schema_element[i].element_helper->match_helper[h].xpath[p] >= 0 ; p++)
    ;
  if (p < schema_element[i].element_helper->max_xpath_seq)
    schema_element[i].element_helper->match_helper[h].xpath[p] = branch;
}


int add_helper_elements(i, i_subr, h)
int i;
int i_subr;
int h;
{
  /* Add the entire helper array of element i_subr to the helper array of element i starting from h */
  int h_subr, p, p_subr;

  /* init scratch mem */
  for (p = 0 ; p < max_xpath_seq ; p++)
    xpath_prefix[p] = -1;
  /* preserve the current xpath that we will work on */
  for (p = 0 ; p < schema_element[i].element_helper->max_xpath_seq ; p++)
    xpath_prefix[p] = schema_element[i].element_helper->match_helper[h].xpath[p];
  for (h_subr = 0 ; h_subr < schema_element[i_subr].element_helper->nxpaths && h < schema_element[i].element_helper->nxpaths ; h_subr++)
  {
    for (p = 0 ; p < schema_element[i].element_helper->max_xpath_seq && xpath_prefix[p] >= 0 ; p++)
    {
      schema_element[i].element_helper->match_helper[h].xpath[p] = xpath_prefix[p];
    }
    for (p_subr = 0 ; p_subr < schema_element[i_subr].element_helper->max_xpath_seq && schema_element[i_subr].element_helper->match_helper[h_subr].xpath[p_subr] >= 0 ; p_subr++)
    {
      if (p < schema_element[i].element_helper->max_xpath_seq)
        schema_element[i].element_helper->match_helper[h].xpath[p++] = schema_element[i_subr].element_helper->match_helper[h_subr].xpath[p_subr];
    }
    h++;
  } /* end for/h_subr */

  return(h);
}


int multiply_helper_elements(i, i_subr, h)
int i;
int i_subr;
int h;
{
  /* Multiply the entire helper array of element i_subr to the helper array of element i starting from h */
  int h_subr, h_shift, p, p_subr;

  while ((h - 1) + schema_element[i_subr].element_helper->nxpaths < schema_element[i].element_helper->nxpaths && schema_element[i].element_helper->match_helper[h].xpath[0] >= 0)
  {
    /* init scratch mem */
    for (p = 0 ; p < max_xpath_seq ; p++)
      xpath_prefix[p] = -1;
    /* preserve the current xpath that we will work on */
    for (p = 0 ; p < schema_element[i].element_helper->max_xpath_seq ; p++)
      xpath_prefix[p] = schema_element[i].element_helper->match_helper[h].xpath[p];
    /* shift down space for nxpaths */
    for (h_shift = schema_element[i].element_helper->nxpaths - schema_element[i_subr].element_helper->nxpaths ; h_shift > h ; h_shift--)
    {
      for (p = 0 ; p < schema_element[i].element_helper->max_xpath_seq ; p++)
        schema_element[i].element_helper->match_helper[h_shift + schema_element[i_subr].element_helper->nxpaths - 1].xpath[p] = schema_element[i].element_helper->match_helper[h_shift].xpath[p];
    } /* end for/h_shift */
    /* multiply new xpaths */
    for (h_subr = 0 ; h_subr < schema_element[i_subr].element_helper->nxpaths ; h_subr++)
    {
      for (p = 0 ; p < schema_element[i].element_helper->max_xpath_seq && xpath_prefix[p] >= 0 ; p++)
        schema_element[i].element_helper->match_helper[h].xpath[p] = xpath_prefix[p];
      for (p_subr = 0 ; p_subr < schema_element[i_subr].element_helper->max_xpath_seq && schema_element[i_subr].element_helper->match_helper[h_subr].xpath[p_subr] >= 0 ; p_subr++)
      {
        if (p < schema_element[i].element_helper->max_xpath_seq)
          schema_element[i].element_helper->match_helper[h].xpath[p++] = schema_element[i_subr].element_helper->match_helper[h_subr].xpath[p_subr];
      }
      h++;
    } /* end for/h_subr */
  } /* end while */

  return(h);
}


void calc_field_match_constraints()
{
  /* iterate through element helpers and calculate field match constraints */
  int i, h;

  for (i = schema_match_start_level ; i < schema_parent.nelements ; i++)
  {
    for (h = 0 ; h < schema_element[i].element_helper->nxpaths ; h++)
    {
      p_curr_xpath = 0;
      traverse_helper_paths(i, h, i, FALSE);
    } /* end for/h */
  } /* end for/i */
}


void traverse_helper_paths(i_helper, h, i, flag_child_of_match_none)
int i_helper;
int h;
int i;
int flag_child_of_match_none;
{
  /* follow the xpaths of the specified helper (i_helper, h) and build its field match constraints */
  int i_subr, j, k;

  j = schema_element[i_helper].element_helper->match_helper[h].xpath[p_curr_xpath];
  for (k = 0 ; k < schema_element[i].record[j]->nfields ; k++)
  {
    switch (schema_element[i].record[j]->field[k].match_control)
    {
      case MATCH_ONE:
             if (schema_element[i].record[j]->field[k].type == FIELD_CDATA)
             {
               if (!flag_child_of_match_none)
               {
                 schema_element[i_helper].element_helper->match_helper[h].min++;
                 schema_element[i_helper].element_helper->match_helper[h].last++;
               }
               if (schema_element[i_helper].element_helper->match_helper[h].max >= 0)
                 schema_element[i_helper].element_helper->match_helper[h].max++;
             }
             break; /* break switch */

      case MATCH_ONE_OR_MANY:
             if (schema_element[i].record[j]->field[k].type == FIELD_CDATA)
               if (!flag_child_of_match_none)
                 schema_element[i_helper].element_helper->match_helper[h].min++;
             schema_element[i_helper].element_helper->match_helper[h].max = -1; /* negative value = infinite */
             schema_element[i_helper].element_helper->match_helper[h].last = 0;
             break; /* break switch */

      case MATCH_NONE_OR_ONE:
             if (schema_element[i].record[j]->field[k].type == FIELD_CDATA)
             {
               if (schema_element[i_helper].element_helper->match_helper[h].max >= 0)
                 schema_element[i_helper].element_helper->match_helper[h].max++;
             }
             else
               flag_child_of_match_none = TRUE;
             schema_element[i_helper].element_helper->match_helper[h].last = 0;
             break; /* break switch */

      case MATCH_NONE_OR_MANY:
             schema_element[i_helper].element_helper->match_helper[h].max = -1; /* negative value = infinite */
             schema_element[i_helper].element_helper->match_helper[h].last = 0;
             if (schema_element[i].record[j]->field[k].type == FIELD_SUBRECORD)
               flag_child_of_match_none = TRUE;
             break; /* break switch */
    }
    if (schema_element[i].record[j]->field[k].type == FIELD_SUBRECORD)
    {
      for (i_subr = schema_match_start_level ; i_subr < schema_parent.nelements ; i_subr++)
      {
        if (xstrcmp(schema_element[i].record[j]->field[k].name, schema_element[i_subr].name) == 0)
        {
          p_curr_xpath++;
          traverse_helper_paths(i_helper, h, i_subr, flag_child_of_match_none);
          flag_child_of_match_none = FALSE;
          break; /* break for/i_subr */
        }
      } /* end for/i_subr */
    }
  } /* end for/k */
}


void debug_print_helper_xpaths(msg, i_helper, h_helper)
char *msg;
int i_helper;
int h_helper;
{
  int i, h, p;

  if (i_helper < 0 && h_helper < 0)
  {
    fprintf(stderr, "DEBUG: start function: debug_print_helper_xpaths%s", newline);
    fprintf(stderr, "-------------------------------%s", newline);
  }
  
  for (i = schema_match_start_level ; i < schema_parent.nelements ; i++)
  {
    for (h = 0 ; h < schema_element[i].element_helper->nxpaths ; h++)
    {
      if ((i_helper < 0 && h_helper < 0) || (i_helper == i && h_helper == h))
      {
        if (i_helper == i && h_helper == h)
          fprintf(stderr, "DEBUG: %s ", msg);
        fprintf(stderr, "%s.h[%d/%d] ", schema_element[i].name, h, schema_element[i].element_helper->nxpaths - 1);
        fprintf(stderr, "min=%d ", schema_element[i].element_helper->match_helper[h].min);
        if (schema_element[i].element_helper->match_helper[h].max < 0)
          fprintf(stderr, "max=INF ");
        else
          fprintf(stderr, "max=%d ", schema_element[i].element_helper->match_helper[h].max);
        fprintf(stderr, "last=%d xpath=( ", schema_element[i].element_helper->match_helper[h].last);
        for (p = 0 ; p < schema_element[i].element_helper->max_xpath_seq ; p++)
          if (schema_element[i].element_helper->match_helper[h].xpath[p] >= 0)
            fprintf(stderr, "%d ", schema_element[i].element_helper->match_helper[h].xpath[p]);
          else
            fprintf(stderr, ". ");
        fprintf(stderr, ")%s", newline);
      }
    }
    if (i_helper < 0 && h_helper < 0)
      fprintf(stderr, "-------------------------------%s", newline);
  }

  if (i_helper < 0 && h_helper < 0)
    fprintf(stderr, "DEBUG: end function: debug_print_helper_xpaths%s", newline);
}
