/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "schema.h"

/* function templates */
void load_rnc();
void create_element_record_rnc();


void load_rnc()
{
  /* Read the RELAX NG Compact file and parse it into relevant structure members.
   * There is plenty of scope for improvement in this subroutine as
   * parsing routines are notoriously messy in C.
   *
   * Any comments that may be encountered are simply filtered out 
   * of the processing.
   *
   * The full RELAX NG Compact functionality is not currently supported.
   */
  long i, lead_trim;
  unsigned char c;

  alloc_tmp_field();
  tmp_nfields             = 0;
  flag_comments           = FALSE;
  flag_char               = FALSE;
  schema_parent.nelements = 0;
  schema_parent.nrecords  = 0;
  schema_parent.nlevels   = 0;
  start_element_name      = NULL;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: load_rnc%s", newline);

  if ((schema_fd = open((char *) schema_filename, O_RDONLY)) > 0)
  {
    utf_getchar_init(schema_fd);
    while (!utf_flag_EOF)
    {
      flag_char            = FALSE;
      input_record.length  = 0;
      tmp_nfields          = 0;

      /* read whole line into array */
      while ((c = utf_getchar(schema_fd)) != '\n' && !utf_flag_EOF)
      {
        /* ignore carriage returns */
        if (c == '\r') 
          continue;

        input_record.line[input_record.length++] = c;

        /* mem checks */
        if (input_record.length > memseg_input_record)
          memadd_input_record();
      }

      /* don't process EOF */
      if (utf_flag_EOF)
        break;

      /* remove leading trim */
      lead_trim = 0;
      while(lead_trim < input_record.length && isspace(input_record.line[lead_trim]))
        lead_trim++;
      if (lead_trim)
      {
        input_record.length -= lead_trim;
        for (i = 0 ; i < input_record.length ; i++)
          input_record.line[i] = input_record.line[i + lead_trim];
      }
      input_record.line[input_record.length] = '\0';

      /* purge blank lines */
      if (input_record.length == 0)
        continue;

      /* check if args also specified inside schema file */
      schema_args();

      /* purge comment lines */
      if (input_record.length > 1 && xstrncmp((unsigned char *) "##", input_record.line, 2) == 0)
      {
        if (sw_debug)
          fprintf(stderr, "DEBUG: RNC: Ignoring comment line: %s%s", input_record.line, newline);
        continue;
      }

      /* break line into fields */
      for (i = 0 ; i < input_record.length ; i++)
      {
        if (! isspace(input_record.line[i])
            && input_record.line[i] != ',' && input_record.line[i] != '|' && input_record.line[i] != '=') 
        {
          if (!flag_char) 
            tmp_field[tmp_nfields++].value = &input_record.line[i];
          flag_char = TRUE;
        }
        else
        {
          if (flag_char) 
            input_record.line[i] = '\0';
          flag_char = FALSE;
        } 

        /* mem checks */
        if (tmp_nfields > memseg_tmp_nfields)
          memadd_tmp_nfields();
      }

      if (tmp_nfields == 2 && xstrcmp(tmp_field[0].value, (unsigned char *) "start") == 0)
      {
        /* Load up parent line */
        start_element_name = (unsigned char *) malloc((xstrlen(tmp_field[1].value) + 1) * sizeof(unsigned char));
        if (start_element_name == NULL) out_of_mem();
        xstrcpy(start_element_name, tmp_field[1].value);
      }
      else if (tmp_nfields > 1 && xstrcmp(tmp_field[1].value, (unsigned char *) "element") == 0)
      {
        /* Load up element lines */
        create_element_record_rnc();
      }
      else if (sw_debug)
      {
        fprintf(stderr, "DEBUG: RNC: Ignoring unsupported schema line: ");
        for (i = 0 ; i < tmp_nfields ; i++)
          fprintf(stderr, "%s ", tmp_field[i].value);
        fprintf(stderr, "%s", newline);
      }

    }
    (void) close(schema_fd);

    if (start_element_name != NULL)
    {
      create_elements();
      if (err_status == 0) build_element_hierachy();
      if (err_status == 0) helper_calc_tree_paths();
    }
    else
    {
      fprintf(stderr, "%s: ERROR: No start element defined in RNC file %s%s", progname, schema_filename, newline);
      err_status = -2;
    }
  }
  else
  {
    fprintf(stderr, "%s: ERROR: Cannot open RNC file %s%s", progname, schema_filename, newline);
    err_status = -2;
  }

  /* cleanup */
  if (start_element_name != NULL)
    (void) free(start_element_name); 
  (void) free(tmp_field);

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: load_rnc%s", newline);
}


void create_element_record_rnc()
{
  /* Populate the schema_record and schema_field structures */
  long j, k, l;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: create_element_record%s", newline);

  /* Filter out elements that xmlfy currently does not support */
  /* don't process element records containing contentspec of { text } */
  k = 0;
  j = xstrlen(tmp_field[2].value);
  if (tmp_nfields > 2)
  {
    while (k < j && tmp_field[2].value[k] != '{')
      k++;
  }
  if ((tmp_nfields > 5 && tmp_field[3].value[0] == '{' && xstrncmp(tmp_field[4].value, (unsigned char *) "text", 4) == 0 && tmp_field[5].value[0] == '}')
      || (tmp_nfields > 4 && k < j && tmp_field[2].value[k] == '{' && xstrncmp(tmp_field[3].value, (unsigned char *) "text", 4) == 0 && tmp_field[4].value[0] == '}')
      || (tmp_nfields > 3 && xstrncmp(tmp_field[3].value, (unsigned char *) "{text}", 6) == 0)
      || (tmp_nfields > 2 && k < j && xstrncmp(&tmp_field[2].value[k], (unsigned char *) "{text}", 6) == 0))
  {
    if (sw_debug)
    {
      fprintf(stderr, "DEBUG: RNC: ignoring \"{ text }\" element definition.%s", newline);
      fprintf(stderr, "DEBUG: end function: create_element_record%s", newline);
    }
    return;
  }

  if (schema_parent.nrecords == 0)
  {
    schema_record = (struct_schema_record *) malloc(sizeof(struct_schema_record));
    if (schema_record == NULL) out_of_mem();
  }
  else
  {
    schema_record = (struct_schema_record *) realloc(schema_record, (schema_parent.nrecords + 1) * sizeof(struct_schema_record));
    if (schema_record == NULL) out_of_mem();
  }

  schema_record[schema_parent.nrecords].name = (unsigned char *) malloc((xstrlen(tmp_field[2].value) + 1) * sizeof(unsigned char));
  if (schema_record[schema_parent.nrecords].name == NULL) out_of_mem();
  xstrcpy(schema_record[schema_parent.nrecords].name, tmp_field[2].value);
  for (j = 0 ; j < xstrlen(schema_record[schema_parent.nrecords].name) ; j++)
  {
    if (schema_record[schema_parent.nrecords].name[j] == '{')
    {
      schema_record[schema_parent.nrecords].name[j] = '\0';
      break; /* for/i */
    }
  }
  schema_record[schema_parent.nrecords].record_count = 0;
  schema_record[schema_parent.nrecords].nfields = 0;

  /* 2 is derived from the following layouts:
   * Expected: <name>(0) element(1) <name>(2) {(3) ... }(4)
   * Crunched: <name>(0) element(1) <name>(2) {...}(3)
   * Squashed: <name>(0) element(1) <name>{...}(2)
   */
  if (tmp_nfields > 2)
  {
    schema_record[schema_parent.nrecords].field = (struct_schema_field *) malloc((tmp_nfields - 2) * sizeof(struct_schema_field));
    if (schema_record[schema_parent.nrecords].field == NULL) out_of_mem();
    /* set kl markers */
    j = 0;
    while (input_record.line[j] != '{' && j < input_record.length) j++; /* move to first open bracket */
  
    while (j < input_record.length)
    {
      while (!is_legal_xml_element_name_char(input_record.line[j]) && j < input_record.length) j++; /* move past all noise chars */
      k = j; /* mark the start of field */
      while (is_legal_xml_element_name_char(input_record.line[j]) && j < input_record.length) j++; /* mark the end of field */
      if (j - k > 0)
      {
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name = (unsigned char *) malloc((j - k + 1) * sizeof(unsigned char));
        if (schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name == NULL) out_of_mem();
        for (l = 0 ; l < j - k ; l++)
          schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name[l] = input_record.line[k + l];
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name[l] = '\0';
  
        /* field match_control */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].match_control = match_control_token(input_record.line[j]); 
  
        /* set initial field type until further examination */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].type = FIELD_CDATA; 
        /* set initial field count */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].field_count = 0; 
  
        schema_record[schema_parent.nrecords].nfields += 1;
      }
      j++;
    }
  
    if (sw_debug)
    {
      fprintf(stderr, "DEBUG: RNC: record(%d) = %s%s", schema_parent.nrecords, schema_record[schema_parent.nrecords].name, newline);
      for (j = 0 ; j < schema_record[schema_parent.nrecords].nfields ; j++)
        fprintf(stderr, "DEBUG: RNC:   field(%ld) = %s match_control=%d%s", j, schema_record[schema_parent.nrecords].field[j].name, schema_record[schema_parent.nrecords].field[j].match_control, newline);
    }

    schema_parent.nrecords++;
  }

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: create_element_record%s", newline);
}
