/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "schema.h"

unsigned char **xsd_file, **xsd_file_orig;
long xsd_lines, memseg_xsd_lines;
typedef struct {
  unsigned char *e_name; /* element name */
  unsigned char *t_name; /* type name */
} struct_et_map; /* element and type map */
struct_et_map *et_map; /* element and type map */
long n_et_map;
unsigned char **ct_names; /* Array defining the complexType names */
long n_ct_names;
unsigned char *attrib_value;
long memseg_attrib_value_length;

/* function templates */
void load_xsd();
void process_xsd_file();
void create_element_record_xsd(int nchoices);
int  get_attrib_value(int line, unsigned char *attrib);
int  xsd_command_equals(int tag_type, char *prefix, char *s1, unsigned char *s2);
void add_xsd_line();
void refresh_xsd_file();
void free_xsd_work();


void load_xsd()
{
  /* Read the XSD file and parse it into relevant structure members.
   * There is plenty of scope for improvement in this subroutine as
   * parsing routines are notoriously messy in C.
   *
   * Any comments that may be encountered are simply filtered out 
   * of the processing.
   *
   * The full XSD functionality is not currently supported.
   */
  long i, lead_trim;
  unsigned char c;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: load_xsd%s", newline);

  alloc_tmp_field();
  tmp_nfields             = 0;
  flag_comments           = FALSE;
  schema_parent.nelements = 0;
  schema_parent.nrecords  = 0;
  schema_parent.nlevels   = 0;
  start_element_name      = NULL;

  /* initialise xsd scratch mem */
  xsd_lines        = 0;
  memseg_xsd_lines = 100;
  xsd_file = (unsigned char **) malloc(memseg_xsd_lines * sizeof(unsigned char *));
  if (xsd_file == NULL) out_of_mem();
  xsd_file_orig = (unsigned char **) malloc(memseg_xsd_lines * sizeof(unsigned char *));
  if (xsd_file_orig == NULL) out_of_mem();
  /* initialise attribute value memory */
  memseg_attrib_value_length = 100;
  attrib_value = (unsigned char *) malloc(memseg_attrib_value_length * sizeof(unsigned char));
  if (attrib_value == NULL) out_of_mem();
  
  if ((schema_fd = open((char *) schema_filename, O_RDONLY)) > 0)
  {
    utf_getchar_init(schema_fd);
    while (!utf_flag_EOF)
    {
      input_record.length  = 0;

      /* read line into array */
      while ((c = utf_getchar(schema_fd)) != '>' && !utf_flag_EOF)
      {
        /* ignore carriage returns and new lines */
        if (c == '\n' || c == '\r') 
          continue;
          
        /* convert all other whitespace to space for ease of comparison */
        if (isspace(c))
          c = ' ';

        input_record.line[input_record.length++] = c;

        /* mem checks */
        if (input_record.length > memseg_input_record)
          memadd_input_record();
      }

      /* don't process EOF */
      if (utf_flag_EOF)
        break;

      /* put back the loop terminator */
      if (c == '>')
        input_record.line[input_record.length++] = c;

      /* remove leading trim */
      lead_trim = 0;
      while(lead_trim < input_record.length && isspace(input_record.line[lead_trim]))
        lead_trim++;
      if (lead_trim)
      {
        input_record.length -= lead_trim;
        for (i = 0 ; i < input_record.length ; i++)
          input_record.line[i] = input_record.line[i + lead_trim];
      }
      input_record.line[input_record.length] = '\0';

      /* check if args also specified inside schema file */
      schema_args();

      /* purge comments and remove blank lines */
      if (purge_comments_blanklines(UTF_8) == 0)
        continue;

      /* ignore lines beginning with <? like <?xml because some DTD writers wrongfully include it */
      if (input_record.length > 1 && xstrncmp((unsigned char *) "<?", input_record.line, 2) == 0)
      {
        if (sw_debug)
          fprintf(stderr, "DEBUG: XSD: Ignoring unsupported schema line: %s%s", input_record.line, newline);
        continue;
      }

      /* ignore lines beginning with <!DOCTYPE because some DTD writers wrongfully include it */
      if (input_record.length > 8 && xstrncmp((unsigned char *) "<!DOCTYPE", input_record.line, 9) == 0)
      {
        if (sw_debug)
          fprintf(stderr, "DEBUG: XSD: Ignoring unsupported schema line: %s%s", input_record.line, newline);
        continue;
      }

      /* Add XML tag to scratch memory */
      add_xsd_line();
    }
    (void) close(schema_fd);

    process_xsd_file();

    create_elements();
    if (err_status == 0) build_element_hierachy();
    if (err_status == 0) helper_calc_tree_paths();
  }
  else
  {
    fprintf(stderr, "%s: ERROR: Cannot open XSD file %s%s", progname, schema_filename, newline);
    err_status = -2;
  }

  /* cleanup */
  free_xsd_work();
  if (start_element_name != NULL)
    (void) free(start_element_name); 
  (void) free(tmp_field); 

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: load_xsd%s", newline);
}


void process_xsd_file()
{
  /* Process the xsd file which is now filtered, in memory and
   * can be read multiple times. */
  long i, j, k, n, line_length;
  int min_occurs, max_occurs;
  char xsd_prefix[10];
  int cntr_choice;
  int flag_match, flag_sequence_tag, flag_choice_tag;

  /* init */
  flag_sequence_tag = FALSE;
  flag_choice_tag   = FALSE;
  cntr_choice       = 0;

  /* Step 1: Load a schema prefix */
  flag_match = FALSE;
  xsd_prefix[0] = '\0'; /* set a default of no schema prefix */
  for (n = 0 ; !flag_match && n < xsd_lines ; n++)
  {
    line_length = xstrlen(xsd_file[n]);
    for (i = 0 ; !flag_match && i < line_length ; i++)
    {
      if (xstrncmp(&xsd_file[n][i], (unsigned char *) ":schema", 7) == 0)
      {
        if (i > 1 && i < 10)
        {
          for (j = 0 ; j < i ; j++)
            xsd_prefix[j] = xsd_file[n][j + 1];
          xsd_prefix[j] = '\0';
        }
        flag_match = TRUE;
      }
    } /* end for/i */
  } /* end for/n */

  /* Step 2: Load targetNamespace */
  for (n = 0 ; n < xsd_lines ; n++)
  {
    if (xsd_command_equals(TAG_START, xsd_prefix, "schema", xsd_file[n]))
    {
      if (get_attrib_value(n, (unsigned char *) " targetNamespace="))
      {
        schema_xsd_target_ns = (unsigned char *) malloc((xstrlen(attrib_value) + 1) * sizeof(unsigned char));
        if (schema_xsd_target_ns == NULL) out_of_mem();
        xstrcpy(schema_xsd_target_ns, attrib_value);

        break; /* for/n */
      }
    }
  } /* end for/n */

  /* Step 3: Identify the parent or root element name - i.e. the first element definition */
  for (n = 0 ; n < xsd_lines ; n++)
  {
    if (xsd_command_equals(TAG_START, xsd_prefix, "element", xsd_file[n])) 
    {
      if (get_attrib_value(n, (unsigned char *) " name="))
      {
        start_element_name = (unsigned char *) malloc((xstrlen(attrib_value) + 1) * sizeof(unsigned char));
        if (start_element_name == NULL) out_of_mem();
        xstrcpy(start_element_name, attrib_value);

        break; /* for/n */
      }
    }
  } /* end for/n */

  /* Step 4: Identify the name for all complexType definitions */
  n_ct_names = 0;
  for (n = 0 ; n < xsd_lines ; n++)
  {
    if (xsd_command_equals(TAG_START, xsd_prefix, "complexType", xsd_file[n]))
    {
      if (get_attrib_value(n, (unsigned char *) " name="))
      {
        if (n_ct_names)
          ct_names = (unsigned char **) realloc(ct_names, (n_ct_names + 1) * sizeof(unsigned char *));
        else
          ct_names = (unsigned char **) malloc((n_ct_names + 1) * sizeof(unsigned char *));
        if (ct_names == NULL) out_of_mem();
        ct_names[n_ct_names] = (unsigned char *) malloc((xstrlen(attrib_value) + 1) * sizeof(unsigned char));
        if (ct_names[n_ct_names] == NULL) out_of_mem();

        xstrcpy(ct_names[n_ct_names], attrib_value);
        n_ct_names++;
      } 
    }
  } /* end for/n */

  /* Step 5: Create a map between element name and complexType name sorted by complexType */
  n_et_map = 0;
  for (i = 0 ; i < n_ct_names ; i++)
  {
    for (n = 0 ; n < xsd_lines ; n++)
    {
      flag_match = FALSE;
      if (xsd_command_equals(TAG_START, xsd_prefix, "element", xsd_file[n])) 
      {
        if (get_attrib_value(n, (unsigned char *) " type="))
        {
          if (xstrcmp(ct_names[i], attrib_value) == 0)
          {
            if (get_attrib_value(n, (unsigned char *) " name="))
            {
              flag_match = TRUE;
              /* ensure unique map names */
              for (j = 0 ; j < n_et_map ; j++)
                if (xstrcmp(et_map[j].t_name, ct_names[i]) == 0 && xstrcmp(et_map[j].e_name, attrib_value) == 0)
                  flag_match = FALSE;

              if (flag_match)
              {
                if (n_et_map)
                  et_map = (struct_et_map *) realloc(et_map, (n_et_map + 1) * sizeof(struct_et_map));
                else
                  et_map = (struct_et_map *) malloc((n_et_map + 1) * sizeof(struct_et_map));
                if (et_map == NULL) out_of_mem();

                et_map[n_et_map].t_name = ct_names[i];
                et_map[n_et_map].e_name = (unsigned char *) malloc((xstrlen(attrib_value) + 1) * sizeof(unsigned char));
                if (et_map[n_et_map].e_name == NULL) out_of_mem();
                xstrcpy(et_map[n_et_map].e_name, attrib_value);

                n_et_map++;
              }
            }
          }
        }
      }
    } /* end for/n */
  } /* end for/i */
  
  /* Step 6: Create records from each complexType */
  for (i = 0 ; i < n_ct_names ; i++)
  {
    tmp_nfields = 0;
    /* load the complexType */
    for (n = 0 ; n < xsd_lines ; n++)
    {
      if (xsd_command_equals(TAG_START, xsd_prefix, "complexType", xsd_file[n]))
      {
        if (get_attrib_value(n, (unsigned char *) " name="))
        {
          if (xstrcmp(ct_names[i], attrib_value) == 0)
          {
            /* found a matching complexType */
            tmp_nfields       = 1;
            flag_sequence_tag = FALSE;
            flag_choice_tag   = FALSE;
            cntr_choice       = 0;
            for (j = n + 1 ; j < xsd_lines ; j++)
            {
              if (xsd_command_equals(TAG_END, xsd_prefix, "complexType", xsd_file[j]))
              {
                n = j;
                break; /* for/j */
              } 
              else if (xsd_command_equals(TAG_START, xsd_prefix, "element", xsd_file[j]))
              {
                flag_match = FALSE;
                k = 8; /* approximate of starting point */
                line_length =  xstrlen(xsd_file[j]);
                while (k < line_length)
                {
                  if (xstrncmp(&xsd_file[j][k], (unsigned char *) " name=", 6) == 0)
                  {
                    k += 7;
                    tmp_field[tmp_nfields].value = &xsd_file[j][k];
                    flag_match = TRUE;
                  }
                  else if (xstrncmp(&xsd_file[j][k], (unsigned char *) " ref=", 5) == 0)
                  {
                    k += 6;
                    tmp_field[tmp_nfields].value = &xsd_file[j][k];
                    flag_match = TRUE;
                  }
                  if (flag_match && (xsd_file[j][k] == '\"' ||  xsd_file[j][k] == '\''))
                  {
                    /* Now get wildcard info */
                    min_occurs = max_occurs = 1;
                    if (get_attrib_value(j, (unsigned char *) " maxOccurs="))
                      if (xstrcmp(attrib_value, (unsigned char *) "unbounded") == 0 || atoi((char *) attrib_value) > 1)
                        max_occurs = 2;
                    if (get_attrib_value(j, (unsigned char *) " minOccurs="))
                      if (xstrcmp(attrib_value, (unsigned char *) "0") == 0)
                        min_occurs = 0;

                    if (min_occurs == 0 && max_occurs == 1)
                      tmp_field[tmp_nfields].match_control =  MATCH_NONE_OR_ONE;
                    else if (min_occurs == 0 && max_occurs > 1)
                      tmp_field[tmp_nfields].match_control =  MATCH_NONE_OR_MANY;
                    else if (min_occurs == 1 && max_occurs > 1)
                      tmp_field[tmp_nfields].match_control =  MATCH_ONE_OR_MANY;
                    else
                      tmp_field[tmp_nfields].match_control =  MATCH_ONE;

                    if (flag_choice_tag)
                    {
                      if (!flag_sequence_tag && tmp_nfields > 1)
                        cntr_choice++;
                      tmp_field[tmp_nfields].choice = cntr_choice;
                    }
                    else
                      tmp_field[tmp_nfields].choice = 0;

                    tmp_nfields++;
                    /* all string processing for this line should be done so ensure end-of-string for tmp_field */
                    xsd_file[j][k] = '\0';

                    /* if element tag is a structure, skip its contents, the last phase will pick it up */
                    if (xsd_file[j][line_length - 2] != '/' && xsd_file[j][line_length - 1] == '>')
                      while(++j < xsd_lines && !xsd_command_equals(TAG_END, xsd_prefix, "element", xsd_file[j]))
                        n++;

                    /* mem checks */
                    if (tmp_nfields > memseg_tmp_nfields)
                      memadd_tmp_nfields();

                    break; /* while */
                  }
                  k++;
                } /* end while */
              }
              else if (xsd_command_equals(TAG_START, xsd_prefix, "sequence", xsd_file[j]))
              {
                flag_sequence_tag = TRUE;
                if (flag_choice_tag && tmp_nfields > 1)
                  cntr_choice++;
              }
              else if (xsd_command_equals(TAG_END, xsd_prefix, "sequence", xsd_file[j]))
                flag_sequence_tag = FALSE;
              else if (xsd_command_equals(TAG_START, xsd_prefix, "choice", xsd_file[j]))
                flag_choice_tag = TRUE;
              else if (xsd_command_equals(TAG_END, xsd_prefix, "choice", xsd_file[j]))
                flag_choice_tag = FALSE;
            } /* end for/j */
          }
        }
      }
    } /* end for/n */

    /* allocate records for each element of this complexType */
    if (tmp_nfields > 0)
    {
      for (n = 0 ; n < n_et_map ; n++)
      {
        if (xstrcmp(et_map[n].t_name, ct_names[i]) == 0)
        {
          tmp_field[0].value         = et_map[n].e_name;
          tmp_field[0].match_control = MATCH_ONE;
          create_element_record_xsd(cntr_choice);
        }
      } /* end for/n */
      refresh_xsd_file();
    }
  } /* end for/i */

  /* Step 7: Scan for structured elements that are NOT of complexType */
  tmp_nfields = 0;
  for (n = 0 ; n < xsd_lines ; n++)
  {
    if (xsd_command_equals(TAG_START, xsd_prefix, "element", xsd_file[n]))
    {
      i = xstrlen(xsd_file[n]);
      if (xsd_file[n][i - 2] != '/' && xsd_file[n][i - 1] == '>')
      {
        flag_match = FALSE;
        if (get_attrib_value(n, (unsigned char *) " type="))
        {
          /* ensure that it is not a complexType */
          for (j = 0 ; j < n_et_map ; j++)
            if (xstrcmp(ct_names[j], attrib_value) == 0)
              flag_match = TRUE;
        }

        if (!flag_match)
        {
          if (get_attrib_value(n, (unsigned char *) " name="))
          {
            j = 0;
            while (j < i && xstrncmp(&xsd_file[n][j], (unsigned char *) " name=", 6) != 0)
              j++;
            j += 7;
            tmp_field[0].value = &xsd_file[n][j];
            while (j < i && (xsd_file[n][j] != '\"' || xsd_file[n][j] != '\"'))
              j++;
            if (j < i)
              xsd_file[n][j] = '\0';
            tmp_field[1].match_control = MATCH_ONE;

            tmp_nfields       = 1;
            flag_sequence_tag = FALSE;
            flag_choice_tag   = FALSE;
            cntr_choice       = 0;
            for (j = n + 1 ; j < xsd_lines ; j++)
            {
              if (xsd_command_equals(TAG_END, xsd_prefix, "element", xsd_file[j]))
              {
                if (tmp_nfields > 1)
                  create_element_record_xsd(cntr_choice);
                n = j;
                break; /* for/j */
              }
              else if (xsd_command_equals(TAG_START, xsd_prefix, "element", xsd_file[j]))
              {
                i = xstrlen(xsd_file[j]);
                if (xsd_file[j][i - 2] == '/' && xsd_file[j][i - 1] == '>')
                {
                  flag_match = FALSE;
                  k = 8; /* approximate of starting point */
                  while (k < xstrlen(xsd_file[j]))
                  {
                    if (xstrncmp(&xsd_file[j][k], (unsigned char *) " name=", 6) == 0)
                    {
                      k += 7;
                      tmp_field[tmp_nfields].value = &xsd_file[j][k];
                      flag_match = TRUE;
                    }
                    else if (xstrncmp(&xsd_file[j][k], (unsigned char *) " ref=", 5) == 0)
                    {
                      k += 6;
                      tmp_field[tmp_nfields].value = &xsd_file[j][k];
                      flag_match = TRUE;
                    }
                    if (flag_match && (xsd_file[j][k] == '\"' ||  xsd_file[j][k] == '\''))
                    {
                      /* Now get wildcard info */
                      min_occurs = max_occurs = 1;
                      if (get_attrib_value(j, (unsigned char *) " maxOccurs="))
                        if (xstrcmp(attrib_value, (unsigned char *) "unbounded") == 0 || atoi((char *) attrib_value) > 1)
                          max_occurs = 2;
                      if (get_attrib_value(j, (unsigned char *) " minOccurs="))
                        if (xstrcmp(attrib_value, (unsigned char *) "0") == 0)
                          min_occurs = 0;

                      if (min_occurs == 0 && max_occurs == 1)
                        tmp_field[tmp_nfields].match_control =  MATCH_NONE_OR_ONE;
                      else if (min_occurs == 0 && max_occurs > 1)
                        tmp_field[tmp_nfields].match_control =  MATCH_NONE_OR_MANY;
                      else if (min_occurs == 1 && max_occurs > 1)
                        tmp_field[tmp_nfields].match_control =  MATCH_ONE_OR_MANY;
                      else
                        tmp_field[tmp_nfields].match_control =  MATCH_ONE;

                      if (flag_choice_tag)
                      {
                        if (!flag_sequence_tag && tmp_nfields > 1)
                          cntr_choice++;
                        tmp_field[tmp_nfields].choice = cntr_choice;
                      }
                      else
                        tmp_field[tmp_nfields].choice = 0;

                      tmp_nfields++;
                      /* all string processing for this line should be done so ensure end-of-string for tmp_field */
                      xsd_file[j][k] = '\0';

                      /* mem checks */
                      if (tmp_nfields > memseg_tmp_nfields)
                        memadd_tmp_nfields();

                      break; /* while */
                    }
                    k++;
                  } /* end while */
                }
              }
              else if (xsd_command_equals(TAG_START, xsd_prefix, "sequence", xsd_file[j]))
              {
                flag_sequence_tag = TRUE;
                if (flag_choice_tag && tmp_nfields > 1)
                  cntr_choice++;
              }
              else if (xsd_command_equals(TAG_END, xsd_prefix, "sequence", xsd_file[j]))
                flag_sequence_tag = FALSE;
              else if (xsd_command_equals(TAG_START, xsd_prefix, "choice", xsd_file[j]))
                flag_choice_tag = TRUE;
              else if (xsd_command_equals(TAG_END, xsd_prefix, "choice", xsd_file[j]))
                flag_choice_tag = FALSE;
            } /* end for/j */
          }
        }
      }
    }
  } /* end for/n */
}


void create_element_record_xsd(nchoices)
int nchoices;
{
  /* Populate the schema_record and schema_field structures */
  long i, j;
  int k;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: create_element_record%s", newline);

  /* The record matching algorithm already treats fields of the parent or root
   * element definition as "best fit" which shares a similarity to "choices",
   * therefore reset any choice parameters for this type of element to preserve
   * the known behaviour. */
  if (tmp_nfields && xstrcmp(tmp_field[0].value, start_element_name) == 0)
  {
    nchoices = 0;
    for (i = 1 ; i < tmp_nfields ; i++)
      tmp_field[i].choice = 0;
  }

  for (k = 0 ; k <= nchoices ; k++)
  {
    if (schema_parent.nrecords == 0)
    {
      schema_record = (struct_schema_record *) malloc(sizeof(struct_schema_record));
      if (schema_record == NULL) out_of_mem();
    }
    else
    {
      schema_record = (struct_schema_record *) realloc(schema_record, (schema_parent.nrecords + 1) * sizeof(struct_schema_record));
      if (schema_record == NULL) out_of_mem();
    }

    schema_record[schema_parent.nrecords].name = (unsigned char *) malloc((xstrlen(tmp_field[0].value) + 1) * sizeof(unsigned char));
    if (schema_record[schema_parent.nrecords].name == NULL) out_of_mem();
    xstrcpy(schema_record[schema_parent.nrecords].name, tmp_field[0].value);
    schema_record[schema_parent.nrecords].record_count = 0;
    schema_record[schema_parent.nrecords].nfields = 0;

    schema_record[schema_parent.nrecords].field = (struct_schema_field *) malloc(tmp_nfields * sizeof(struct_schema_field));
    if (schema_record[schema_parent.nrecords].field == NULL) out_of_mem();
    for (i = 1 ; i < tmp_nfields ; i++)
    {
      if (tmp_field[i].choice == k)
      {
        j = xstrlen(tmp_field[i].value);
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name = (unsigned char *) malloc((j + 1) * sizeof(unsigned char));
        if (schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name == NULL) out_of_mem();
        xstrcpy(schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].name, tmp_field[i].value);

        /* field match_control */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].match_control = tmp_field[i].match_control; 
        /* set initial field type until further examination */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].type = FIELD_CDATA; 
        /* set initial field count */
        schema_record[schema_parent.nrecords].field[schema_record[schema_parent.nrecords].nfields].field_count = 0; 

        schema_record[schema_parent.nrecords].nfields += 1;
      }
    } /* end for/i */

    if (sw_debug)
    {
      fprintf(stderr, "DEBUG: Schema: record(%d) = %s%s", schema_parent.nrecords, schema_record[schema_parent.nrecords].name, newline);
      for (j = 0 ; j < schema_record[schema_parent.nrecords].nfields ; j++)
        fprintf(stderr, "DEBUG: Schema:   field(%ld) = %s match_control=%d%s", j, schema_record[schema_parent.nrecords].field[j].name, schema_record[schema_parent.nrecords].field[j].match_control, newline);
    }

    schema_parent.nrecords++;
  } /* end for/k */

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: create_element_record%s", newline);
}


int xsd_command_equals(tag_type, prefix, s1, s2)
int tag_type;
char *prefix;
char *s1;
unsigned char *s2;
{
  /* Find a XSD command */
  int ret;
  char xsd_command[20];

  if (tag_type == TAG_START)
    (void) sprintf(xsd_command, "<%s%s", prefix, s1);
  else
    (void) sprintf(xsd_command, "</%s%s", prefix, s1);

  if (xstrncmp(s2, (unsigned char *) xsd_command, strlen(xsd_command)) == 0)
    ret = TRUE;
  else
    ret = FALSE;

  return(ret);
}


int get_attrib_value(line, attrib)
int line;
unsigned char *attrib;
{
  /* Check if attribute exists inside the specified line and store its value in a common variable */
  long i, j, k;
  int attrib_length, ret;

  ret = FALSE;

  attrib_length = xstrlen(attrib);
  for (i = 0 ; i < xstrlen(xsd_file[line]) ; i++)
  {
    if (xstrncmp(&xsd_file[line][i], attrib, attrib_length) == 0) 
    {
      k = 0;
      for (j = i + attrib_length + 1 ; j < xstrlen(xsd_file[line]) && xsd_file[line][j] != '\"' && xsd_file[line][j] != '\'' ; j++)
        k++;

      /* mem checks */
      if (k > memseg_attrib_value_length)
      {
        memseg_attrib_value_length += k + 100;
        attrib_value = (unsigned char *) realloc(attrib_value, (memseg_attrib_value_length + 1) * sizeof(unsigned char));
      }

      for (j = 0 ; j < k ; j++)
        attrib_value[j] = xsd_file[line][i + attrib_length + 1 + j];
      attrib_value[j] = '\0';
      ret = TRUE;

      break; /* break for/i */
    }
  } /* end for/i */
  return(ret);
}


void add_xsd_line()
{
  /* Add a filtered line to XSD scratch memory */
  long i;

  /* mem checks */
  if (xsd_lines > memseg_xsd_lines)
  {
    memseg_xsd_lines += 100;
    xsd_file = (unsigned char **) realloc(xsd_file, memseg_xsd_lines * sizeof(unsigned char *));
    if (xsd_file == NULL) out_of_mem();
    xsd_file_orig = (unsigned char **) realloc(xsd_file_orig, memseg_xsd_lines * sizeof(unsigned char *));
    if (xsd_file_orig == NULL) out_of_mem();
  }

  /* add line */
  xsd_file[xsd_lines] = (unsigned char *) malloc((input_record.length + 1) * sizeof(unsigned char));
  if (xsd_file[xsd_lines] == NULL) out_of_mem();
  xsd_file_orig[xsd_lines] = (unsigned char *) malloc((input_record.length + 1) * sizeof(unsigned char));
  if (xsd_file_orig[xsd_lines] == NULL) out_of_mem();
  for (i = 0 ; i < input_record.length ; i++)
  {
    xsd_file[xsd_lines][i] = input_record.line[i];
    xsd_file_orig[xsd_lines][i] = input_record.line[i];
  }

  xsd_lines++;
}


void refresh_xsd_file()
{
  /* Copy backed up xsd_file to working copy */
  long n;

  for (n = 0 ; n < xsd_lines ; n++)
    xstrcpy(xsd_file[n], xsd_file_orig[n]);
}


void free_xsd_work()
{
  /* Free XSD scratch memory */
  long i;

  /* free scratch file allocation */
  for (i = 0 ; i < xsd_lines ; i++)
  {
    (void) free(xsd_file[i]);
    (void) free(xsd_file_orig[i]);
  }
  (void) free(xsd_file);
  (void) free(xsd_file_orig);

  /* free ct_names allocation */
  for (i = 0 ; i < n_ct_names ; i++)
    (void) free(ct_names[i]);
  (void) free(ct_names);

  /* free et_map allocation */
  for (i = 0 ; i < n_et_map ; i++)
    (void) free(et_map[i].e_name);
  (void) free(et_map);

  (void) free(attrib_value);
}
