/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "schema.h"

int schema_fd;
int flag_comments, flag_char;
unsigned char *start_element_name; /* for RNC and XSD */
struct_tmp_field *tmp_field;
long tmp_nfields;
long memseg_tmp_nfields;

/* function templates */
void create_elements();
void build_element_hierachy();
void count_schema_levels(int depth, int i);
long purge_comments_blanklines(int encoding);
int  match_control_token(unsigned char c1);
int  is_legal_xml_element_name_char(unsigned char c1);
void alloc_tmp_field();
void memadd_tmp_nfields();
void debug_walk_schema_tree(int level, int curr_record, int nrecord);


void schema()
{
  /* load a schema if any */
  if (sw_schema == SCHEMA_DTD)
    load_dtd();
  else if (sw_schema == SCHEMA_RNC)
    load_rnc();
  else if (sw_schema == SCHEMA_XSD)
    load_xsd();
  else if (sw_debug && sw_schema == SCHEMA_DEFAULT)
    fprintf(stderr, "DEBUG: Using default schema.%s", newline);
}


void create_elements()
{
  /* iterate through records and create elements */
  int i, j, k;
  int flag_add_element;
  struct_schema_element tmp_schema_element;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: create_elements%s", newline);

  schema_parent.nelements = 0;
  schema_parent.name      = NULL;

  for (i = 0 ; i < schema_parent.nrecords ; i++)
  {
    flag_add_element = TRUE;
    /* associate similar named record with existing element */
    for (j = i ; j > 0 && flag_add_element ; j--)
    {
      if (xstrcmp(schema_record[j].name, schema_record[i].name) == 0)
      {
        for (k = 0 ; k < schema_parent.nelements ; k++)
          if (xstrcmp(schema_element[k].name, schema_record[j].name) == 0)
            break;

        if (k < schema_parent.nelements)
        {
          flag_add_element = FALSE;
          schema_element[k].record = (struct_schema_record **) realloc(schema_element[k].record, (schema_element[k].nrecords + 1) * sizeof(struct_schema_record *));
          if (schema_element[k].record == NULL) out_of_mem();
          schema_element[k].record[schema_element[k].nrecords] = &schema_record[j];
          schema_element[k].nrecords++;
        }
      }
    }

    /* add new element and associate the record that triggered it */
    if (flag_add_element)
    {
      if (schema_parent.nelements)
        schema_element = (struct_schema_element *) realloc(schema_element, (schema_parent.nelements + 1) * sizeof(struct_schema_element));
      else
        schema_element = (struct_schema_element *) malloc(sizeof(struct_schema_element));
      if (schema_element == NULL) out_of_mem();

      schema_element[schema_parent.nelements].element_count = 0; 
      schema_element[schema_parent.nelements].element_helper = NULL;
      schema_element[schema_parent.nelements].nrecords = 0;
      schema_element[schema_parent.nelements].record = (struct_schema_record **) malloc(sizeof(struct_schema_record *));
      if (schema_element[schema_parent.nelements].record == NULL) out_of_mem();
      schema_element[schema_parent.nelements].record[0] = &schema_record[i];
      schema_element[schema_parent.nelements].name = schema_record[i].name;
      schema_element[schema_parent.nelements].nrecords++;

      /* define the parent element*/
      if (sw_schema == SCHEMA_DTD && schema_parent.nelements == 0)
      {
        /* make the first element the parent element */
        schema_parent.name = schema_element[0].name;
        schema_parent.record = schema_element[0].record[0];
      }
      else if (sw_schema == SCHEMA_RNC || sw_schema == SCHEMA_XSD)
      {
        /* make the start element the parent element */
        if (xstrcmp(start_element_name, schema_element[schema_parent.nelements].name) == 0)
        {
          schema_parent.name = schema_element[schema_parent.nelements].name;
          schema_parent.record = schema_element[schema_parent.nelements].record[0];
        }
      }
      schema_parent.nelements++;
    }
  }

  if (schema_parent.name == NULL)
  {
    fprintf(stderr, "%s: ERROR: No parent element found in schema file '%s'.%s", progname, schema_filename, newline);
    err_status = -2;
  }

  /* For element matching efficiency, re-arrange parent record (entry point) to be array element 0 */
  if (schema_parent.nelements && schema_parent.name != schema_element[0].name)
  {
    for (i = 1 ; i < schema_parent.nelements ; i++)
    {
      if (schema_parent.name == schema_element[i].name)
      {
        /* swaparoo */
        tmp_schema_element   = schema_element[0];
        schema_element[0]    = schema_element[i];
        schema_element[i]    = tmp_schema_element;
        /* update the parent pointer */
        schema_parent.name   = schema_element[0].name;
        schema_parent.record = schema_element[0].record[0];
        break; /* for/i */
      }
    }
  }


  if (sw_debug)
  {
    fprintf(stderr, "DEBUG: Schema: parent element = \"%s\"%s", schema_parent.name, newline);
    for (i = 0 ; i < schema_parent.nelements ; i++)
    {
      fprintf(stderr, "DEBUG: Schema: element(%d) = %s%s", i, schema_element[i].name, newline);
      for (j = 0 ; j < schema_element[i].nrecords ; j++)
      {
        fprintf(stderr, "DEBUG: Schema:   record(%d) = %s%s", j, schema_element[i].record[j]->name, newline);
        for (k = 0 ; k < schema_element[i].record[j]->nfields ; k++)
        {
          fprintf(stderr, "DEBUG: Schema:     field(%d) = %s%s", k, schema_element[i].record[j]->field[k].name, newline);
        }
      }
    }
  }

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: create_elements%s", newline);
}


void build_element_hierachy()
{
  /* iterate through elements and build the hierachy */
  int i, k;

  for (k = 0 ; k < schema_parent.record->nfields ; k++)
    for (i = 1 ; i < schema_parent.nelements ; i++)
      if (xstrcmp(schema_element[i].name, schema_parent.record->field[k].name) == 0)
        schema_parent.record->field[k].type = FIELD_SUBRECORD;
  /* count schema levels and update fields that are really records */
  for (i = 1 ; i < schema_parent.nelements ; i++)
    count_schema_levels(2, i);
  /* Schema may have more levels than cmd line options so allocate that extra level memory */
  if (schema_parent.nlevels >= nlevels)
    increase_nlevels(schema_parent.nlevels + 1);

  /* allocate that initial schema_xpath memory that we could not allocate until final nlevels is known */
  for (i = 0 ; i < DEFAULT_NFIELDS ; i++)
  {
    input_record.field[i]->schema_xpath = (int *) malloc(nlevels * sizeof(int));
    if (input_record.field[i]->schema_xpath == NULL) out_of_mem();
  }
  
  /* check that element in option --makeparent <elementname> exists */
  if (!schema_match_start_level)
  {
    if (xstrcmp(schema_match_start_name, (unsigned char *) "0") != 0)
    {
      for (i = 0 ; i < schema_parent.nelements ; i++)
        if (xstrcmp(schema_match_start_name, schema_record[i].name) == 0)
          break; /* for/i */
      if (i >= schema_parent.nelements)
      {
        fprintf(stderr, "%s: ERROR: Element record '%s' not found in schema.%s", progname, schema_match_start_name, newline);
        err_status = -1;
      }
    }
  }



  if (sw_debug)
  {
    fprintf(stderr, "DEBUG: Walk schema tree:%s", newline);
    for (i = 0 ; i < schema_parent.nelements ; i++)
    {
      if (schema_parent.name == schema_record[i].name)
      {
        debug_walk_schema_tree(0, i, 0);
        break;
      }
    }
  }
}


void count_schema_levels(depth, i)
int depth;
int i;
{
  /* count schema levels and update fields that are really records */
  int i_subr, j, k;

  if (depth > schema_parent.nlevels)
    schema_parent.nlevels = depth;

  for (j = 0 ; j < schema_element[i].nrecords ; j++)
  {
    for (k = 0 ; k < schema_element[i].record[j]->nfields ; k++)
    {
      /* find matching element name */
      for (i_subr = 1 ; i_subr < schema_parent.nelements ; i_subr++)
      {
        if (i_subr != i && xstrcmp(schema_element[i_subr].name, schema_element[i].record[j]->field[k].name) == 0)
        {
          schema_element[i].record[j]->field[k].type = FIELD_SUBRECORD;
          count_schema_levels(depth + 1, i_subr);
          break;
        }
      } /* end for/i_subr */
    } /* end for/k */
  } /* end for/j */
}
          

int match_control_token(unsigned char c1)
/* unsigned char c1; This approach may generate a compiler warning */
{
  /* set wildcard token */
  int ret;

  if (c1 == '?')
    ret = MATCH_NONE_OR_ONE; 
  else if (c1 == '*')
    ret = MATCH_NONE_OR_MANY; 
  else if (c1 == '+')
    ret = MATCH_ONE_OR_MANY; 
  else
    ret = MATCH_ONE; 

  return (ret);
}


int is_legal_xml_element_name_char(unsigned char c1)
/* unsigned char c1; This approach may generate a compiler warning */
{
  /* check if character is legal for use inside an XML element name */
  int ret;

  if (isalnum(c1))
    ret = TRUE;
  else if (c1 == '-')
    ret = TRUE;
  else if (c1 == '_')
    ret = TRUE;
  else if (c1 == ':')
    ret = TRUE;
  else if (c1 == '.')
    ret = TRUE;
  else if (c1 > 0x7f && c1 < 0xc0) /* UTF-8 */
    ret = TRUE;
  else
    ret = FALSE;

  return (ret);
}


long purge_comments_blanklines(encoding)
int  encoding;
{
  /* remove comments and blank lines from the schema input line */
  long i, j, c_on, c_off;
  int saved_encoding, saved_charwidth;

  c_on  = -1;
  c_off = -1;
  /* adjust global encoding variables for handling multiple encoding
   * situations.
   *   E.g. Schema files are ASCII/UTF-8 and input stream can be any encoding.
   *   E.g. unxml option can be any encoding. */
  saved_encoding  = input_record.encoding;
  saved_charwidth = input_record.charwidth;
  if (encoding != input_record.encoding)
  {
    input_record.encoding = encoding;
    if (input_record.encoding == UTF_8)
      input_record.charwidth = 1;
    else if (input_record.encoding == UTF_16BE || input_record.encoding == UTF_16LE)
      input_record.charwidth = 2;
    else if (input_record.encoding == UTF_32BE || input_record.encoding == UTF_32LE)
      input_record.charwidth = 4;
  }

  for (i = 0 ; i < input_record.length ; i += input_record.charwidth)
  {
    if (utf_char_equals(i, '<'))
      if (utf_char_equals(i + input_record.charwidth, '!'))
        if (utf_char_equals(i + (2 * input_record.charwidth), '-'))
          if (utf_char_equals(i + (3 * input_record.charwidth), '-'))
          {
            c_on = i;
            flag_comments = TRUE;
          }

    if (utf_char_equals(i, '-'))
      if (utf_char_equals(i + input_record.charwidth, '-'))
        if (utf_char_equals(i + (2 * input_record.charwidth), '>'))
        {
          c_off = i + (2 * input_record.charwidth);
          flag_comments = FALSE;
        }

    /* purge everything up to end-of-comment token */
    if (c_off >= 0)
    {
      if (c_on < 0) c_on = 0;
      for (j = c_off + input_record.charwidth ; j <= input_record.length ; j++)
        input_record.line[j - input_record.charwidth - c_off + c_on] = input_record.line[j];
      input_record.length -= (c_off - c_on + input_record.charwidth);
      i = c_off;

      /* reset */
      c_on  = -1;
      c_off = -1;
    }
  }

  if (flag_comments)
  {
    if (c_on < 0)  /* the whole line is a comment so purge it */
      input_record.length = 0;
    else           /* purge everything after start-of-comment token */
      input_record.length = c_on;
  }

  /* purge blank lines */
  for (i = 0 ; i < input_record.length ; i += input_record.charwidth)
  {
    if (!utf_isspace(i))
      break;
  }
  if (i == input_record.length)
    input_record.length = 0;

  /* terminate resized string */
  if (input_record.encoding == UTF_8)
    input_record.line[input_record.length] = '\0';

  /* restore the saved encoding */
  input_record.encoding  = saved_encoding;
  input_record.charwidth = saved_charwidth;

  return(input_record.length);
}


void alloc_tmp_field()
{
  /* allocate initial memory for tmp fields */
  memseg_tmp_nfields = DEFAULT_NFIELDS;
  tmp_field = (struct_tmp_field *) malloc(memseg_tmp_nfields * sizeof(struct_tmp_field));
  if (tmp_field == NULL) out_of_mem();
}


void memadd_tmp_nfields()
{
  /* re-allocate more memory for more tmp fields */
  memseg_tmp_nfields = ((tmp_nfields + MEMSEG_THRESHOLD) / DEFAULT_NFIELDS + 1) * DEFAULT_NFIELDS - MEMSEG_THRESHOLD;
  tmp_field = (struct_tmp_field *) realloc(tmp_field, (memseg_tmp_nfields + MEMSEG_THRESHOLD) * sizeof(struct_tmp_field));
  if (tmp_field == NULL) out_of_mem();
}


void debug_walk_schema_tree(level, curr_record, nrecord)
int level;
int curr_record;
int nrecord;
{
  /* walk the tree and associate levels */
  int i, j, k, p;

  fprintf(stderr, "DEBUG: Schema: ");
  /* padding */
  for (p = 0 ; p < level ; p++)
    fprintf(stderr, "  ");
  fprintf(stderr, "%s[%d]%s", schema_record[curr_record].name, nrecord, newline);

  level++;

  for (i = 0 ; i < schema_record[curr_record].nfields ; i++)
  {
    if (schema_record[curr_record].field[i].type == FIELD_CDATA)
    {
      fprintf(stderr, "DEBUG: Schema: ");
      /* padding */
      for (p = 0 ; p < level ; p++)
        fprintf(stderr, "  ");
      fprintf(stderr, "%s%s", schema_record[curr_record].field[i].name, newline);
    }
    else
    {
      k = 0;
      for (j = 0 ; j < schema_parent.nrecords ; j++)
      {
        if (xstrcmp(schema_record[curr_record].field[i].name, schema_record[j].name) == 0)
        {
          if (level < schema_parent.nlevels) 
            debug_walk_schema_tree(level, j, k++);
        }
      }
    }
  }
}
