/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "unxml.h"

int flag_comments;
long i_pos, t_open, t_close;

/* function templates */
int  is_tag();
void insert_delimiter(int level);
void unescape_chars();
void insert_quote(int level, int quote_action);


void unxml()
{
  /* Do the opposite of xmlfy and remove all that XML bracket racket.
   * Full XML detoxification is not fully supported as this routine
   * focuses primarily on undoing what xmlfy normally does.
   * 
   * In hindsight, and in staying with Un*x tradition, I should have
   * written two tools and called them intox and detox respectively.
   */
  int level, i;

  /* initialise */
  flag_comments = FALSE;
  level = -1;

  /* try and get the encoding being used */
  utf_get_encoding();

  while (!utf_flag_EOF)
  {
    input_record.length = 0;

    do
    {
      input_record.line[input_record.length++] = utf_getchar(STDIN_FD);

      /* mem checks */
      if (input_record.length > memseg_input_record)
        memadd_input_record();
    }
    while (!utf_flag_EOF && (input_record.length % input_record.charwidth || !utf_char_equals_level1(input_record.length - 1, '>')));

    /* don't process EOF */
    if (utf_flag_EOF)
      break;

    /* set close tag marker */
    t_close = input_record.length - input_record.charwidth;

    /* purge comments and remove blank lines */
    if (purge_comments_blanklines(input_record.encoding) == 0)
      continue;

    /* remove XML */
    output_record.length = 0;
    i_pos = 0;
    while (i_pos < input_record.length)
    {
      switch(is_tag())
      {
        case TAG_START:
               /* move to new level */
               level++;
               /* only print delimiter if more than one field */
               if (level >= 0 && level < nlevels)
               {
                 if (output_record.field_cntr[level] > 0)
                 {
                   insert_delimiter(level);
                   if (level == 2)
                     insert_quote(level, QUOTE_OPEN);
                 }
                 else
                 {
                   if (level == 2)
                     insert_quote(level, QUOTE_OPEN); /* open quote for 1st field */
                 }
                 output_record.field_cntr[level]++;
               }
               else
                 insert_delimiter(level);
               break;

        case TAG_END:
               /* clear field counters of next higher level */
               if (level + 1 >= 0 && level + 1 < nlevels)
                 output_record.field_cntr[level + 1] = 0;
               if (level == 2) insert_quote(level, QUOTE_CLOSE);
               /* move to new level */
               level--;
               break;

        case TAG_EMPTY:
        case TAG_PROLOGUE:
        case TAG_DOCTYPE:
               break;

        default:
               if (level >= 0)
               {
                 if (input_record.encoding == UTF_8)
                   output_record.line[output_record.length++] = input_record.line[i_pos++];
                 else
                 {
                   for (i = 0 ; i < input_record.charwidth && i_pos < input_record.length ; i++)
                     output_record.line[output_record.length++] = input_record.line[i_pos++];
                 }
               }
               else
                 i_pos += input_record.charwidth;

               /* mem checks */
               if (output_record.length > memseg_output_record)
                 memadd_output_record(output_record.length);
      }
    }
    unescape_chars();
    utf_print(STDOUT_FD, output_record.line, output_record.length);
  }

  /* Clarification when using the default level 1 delimiter NEWLINE:
   * Most Unix text files finish their last line with a newline character.
   * By default, terminate the very last record with a NEWLINE in order
   * to furnish a proper Unix line.
   * If you really don't want the last NEWLINE character as a record
   * delimiter then use the explicit option -F1 "\n" 
   */
  if (delimiter[1]->ndelimiters == 0)
  {
    strcpy_output_record(0, newline);
    utf_print(STDOUT_FD, output_record.line, output_record.length);
  }
}


int is_tag()
{
  /* Determine the tag type */
  int ret;
  long i, lead_trim;

  ret = FALSE;
  i   = 0;

  if (utf_char_equals(i_pos, '<'))
  {
    t_open = i_pos;

    if (utf_char_equals(t_open + input_record.charwidth, '/'))
      ret = TAG_END;
    else if (t_close >= input_record.charwidth && utf_char_equals(t_close - input_record.charwidth, '/'))
      ret = TAG_EMPTY;
    else if (utf_char_equals(t_open + input_record.charwidth, '?'))
      ret = TAG_PROLOGUE;
    else if (utf_char_equals(t_open + input_record.charwidth, '!'))
      ret = TAG_DOCTYPE;
    else
      ret = TAG_START;

    /* remove leading trim of all tags */
    if (t_open > 0)
    {
      lead_trim = 0;
      for (i = t_open - input_record.charwidth ; i >= 0 && utf_isspace(i) ; i -= input_record.charwidth)
        lead_trim += input_record.charwidth;
      if (output_record.length > lead_trim)
        output_record.length -= lead_trim;
      else
        output_record.length = 0;
    }

    /* update position */
    i_pos = t_close + input_record.charwidth;
  }

  return(ret);
}


void insert_delimiter(level)
int level;
{
  /* insert the delimiter string token between the fields */
  int i;

  if (level < nlevels && level > 0 && delimiter[level]->ndelimiters)
  {
    for (i = 0 ; i < delimiter[level]->utf_length[0] ; i++)
      output_record.line[output_record.length++] = delimiter[level]->utf_d[0][i];
  }
  else if (level > 1)
    strcat_output_record((unsigned char *) " ");
  else 
    strcat_output_record(newline); /* perhaps ambiguous but also use new-line for level 0 */
}


void unescape_chars()
{
  /* reverts escaped XML chars back to their ASCII equivalent */
  long i, j;

  /* scan output_record and adjust length */
  for (i = 0 ; i < output_record.length ; i++)
  {
    if (utf_char_equals_2(i, '&', output_record.line, output_record.length, output_record.encoding))
    {
      if (utf_char_equals_2(i + input_record.charwidth, 'l', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (2 * input_record.charwidth), 't', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (3 * input_record.charwidth), ';', output_record.line, output_record.length, output_record.encoding))
      {
        j = output_record.length;
        output_record.length = i; /* fudge for strcat routine */
        strcat_output_record((unsigned char *) "<");
        output_record.length = j - (3 * input_record.charwidth);
        for (j = i + input_record.charwidth ; j < output_record.length ; j++)
          output_record.line[j] = output_record.line[j + (3 * input_record.charwidth)];
      }
      else if (utf_char_equals_2(i + input_record.charwidth, 'g', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (2 * input_record.charwidth), 't', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (3 * input_record.charwidth), ';', output_record.line, output_record.length, output_record.encoding))
      {
        j = output_record.length;
        output_record.length = i; /* fudge for strcat routine */
        strcat_output_record((unsigned char *) ">");
        output_record.length = j - (3 * input_record.charwidth);
        for (j = i + input_record.charwidth ; j < output_record.length ; j++)
          output_record.line[j] = output_record.line[j + (3 * input_record.charwidth)];
      }
      else if (utf_char_equals_2(i + input_record.charwidth, 'a', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (2 * input_record.charwidth), 'm', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (3 * input_record.charwidth), 'p', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (4 * input_record.charwidth), ';', output_record.line, output_record.length, output_record.encoding))
      {
        j = output_record.length;
        output_record.length = i; /* fudge for strcat routine */
        strcat_output_record((unsigned char *) "&");
        output_record.length = j - (4 * input_record.charwidth);
        for (j = i + input_record.charwidth ; j < output_record.length ; j++)
          output_record.line[j] = output_record.line[j + (4 * input_record.charwidth)];
      }
      else if (utf_char_equals_2(i + input_record.charwidth, 'q', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (2 * input_record.charwidth), 'u', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (3 * input_record.charwidth), 'o', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (4 * input_record.charwidth), 't', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (5 * input_record.charwidth), ';', output_record.line, output_record.length, output_record.encoding))
      {
        j = output_record.length;
        output_record.length = i; /* fudge for strcat routine */
        strcat_output_record((unsigned char *) "\"");
        output_record.length = j - (5 * input_record.charwidth);
        for (j = i + input_record.charwidth ; j < output_record.length ; j++)
          output_record.line[j] = output_record.line[j + (5 * input_record.charwidth)];
      }
      else if (utf_char_equals_2(i + input_record.charwidth, 'b', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (2 * input_record.charwidth), 'r', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (3 * input_record.charwidth), 'v', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (4 * input_record.charwidth), 'b', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (5 * input_record.charwidth), 'a', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (6 * input_record.charwidth), 'r', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (7 * input_record.charwidth), ';', output_record.line, output_record.length, output_record.encoding))
      {
        j = output_record.length;
        output_record.length = i; /* fudge for strcat routine */
        strcat_output_record((unsigned char *) "|");
        output_record.length = j - (7 * input_record.charwidth);
        for (j = i + input_record.charwidth ; j < output_record.length ; j++)
          output_record.line[j] = output_record.line[j + (7 * input_record.charwidth)];
      }
      else if (utf_char_equals_2(i + input_record.charwidth, 'a', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (2 * input_record.charwidth), 'p', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (3 * input_record.charwidth), 'o', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (4 * input_record.charwidth), 's', output_record.line, output_record.length, output_record.encoding)
          && utf_char_equals_2(i + (5 * input_record.charwidth), ';', output_record.line, output_record.length, output_record.encoding))
      {
        j = output_record.length;
        output_record.length = i; /* fudge for strcat routine */
        strcat_output_record((unsigned char *) "\'");
        output_record.length = j - (5 * input_record.charwidth);
        for (j = i + input_record.charwidth ; j < output_record.length ; j++)
          output_record.line[j] = output_record.line[j + (5 * input_record.charwidth)];
      }
    }
  }
}


void insert_quote(level, quote_action)
int level;
int quote_action;
{
  /* Insert quotes for level 2 fields */

  if (sw_field_quoting == TRUE)
  {
    if (quotechar_type == 0)
      strcat_output_record((unsigned char *) "\"");
    else if (quotechar_type == 1)
      strncat_output_record(&quotechars[0], 1);
    else if (quotechar_type == 2)
    {
      if (quote_action == QUOTE_OPEN)
        strncat_output_record(&quotechars[0], 1);
      else
        strncat_output_record(&quotechars[1], 1);
    }
  }
}
