/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "utf_io.h"

/* NOTE: We don't bother with wchar type and its derivatives for handling 
 * UTF-16. Instead we store code points on a per byte basis which gives
 * better flexibility, portability and extensibility e.g. UTF-32 */

/* The variable below is only used to stop compile time FORTIFY_SOURCE
 * warnings when casting a return-value-function to (void).
 * E.g (void) write(...) now becomes write_n = (long) write(...) and write_n
 * is ignored. Could use ssize_t but safer with type long.  */
long write_n;

/* function templates */
void utf_getchar_init(int fd);
unsigned char utf_getchar(int fd);
void utf_print(int fd, unsigned char *buf, long buf_bytes);
int  utf_read_bom(int fd);
void utf_print_bom(int fd);


void utf_getchar_init(fd)
int fd;
{
  /* initialise variables for mimicking char reads from a block read */
  utf_flag_EOF = FALSE;
  utf_read     = 0;
  utf_cntr     = 0;
  if (sw_summary && sw_digest && fd == STDIN_FD)
  {
    MD5Init(&md5_out_context);
    MD5Init(&md5_in_context);
    SHA512Reset(&sha512_in_context);
    SHA512Reset(&sha512_out_context);
  }
}


unsigned char utf_getchar(fd)
int fd;
{
  /* don't read beyond EOF more than once */
  if (utf_flag_EOF)
    return(0);

  /* Read a block of input and return one char at a time */
  if (utf_cntr >= utf_read)
  {
    utf_read = read(fd, utf_buf, BUFSIZE);
    utf_cntr = 0;
    if (utf_read < 1)
      utf_flag_EOF = TRUE;
    else if (sw_summary && sw_digest && fd == STDIN_FD)
    {
      MD5Update(&md5_in_context, utf_buf, utf_read);
      SHA512Input(&sha512_in_context, utf_buf, utf_read);
    }
  }

  return(utf_buf[utf_cntr++]);
}


void utf_print(fd, buf, buf_bytes)
int fd;
unsigned char *buf;
long buf_bytes;
{
  /* print UTF stream */
  /* printf() won't work here because of the need to print the null character */
  long i;

  /* output in manageable block sizes */
  for (i = 0 ; i < buf_bytes / BUFSIZE ; i++)
    write_n = (long) write(fd, &buf[i * BUFSIZE], BUFSIZE);
  i = buf_bytes % BUFSIZE;
  if (i > 0)
    write_n = (long) write(fd, &buf[buf_bytes - i], i);

  if (sw_summary && sw_digest && fd == STDOUT_FD && sw_process_output_digest)
  {
    MD5Update(&md5_out_context, buf, buf_bytes);
    SHA512Input(&sha512_out_context, buf, buf_bytes);
  }
}


int utf_read_bom(fd)
int fd;
{
  /* Pre-load the read block buffer and by-pass the BOM if any. */
  /* This must be the first read and is only called once. 
   * NOTE: encoding found by BOM option overides any user encoding options */
  int flag_bom_found;

  /* initialise */
  utf_getchar_init(fd);
  xstrcpy(input_record.bom_name, (unsigned char *) BOM_NONE_NAME);
  flag_bom_found = FALSE;

  /* Pre-load a block of input stream */
  if (!utf_flag_EOF && utf_cntr >= utf_read)
  {
    utf_read = read(fd, utf_buf, BUFSIZE);
    if (utf_read < 1)
      utf_flag_EOF = TRUE;
    else
    {
      if (utf_read > 2 && xstrncmp((unsigned char *) UTF8_BOM, utf_buf, 3) == 0)
      {
        flag_bom_found = TRUE;
        sw_encoding = UTF_8;
        xstrcpy(input_record.bom_name, (unsigned char *) UTF_8_NAME);
        xstrcpy(output_record.encoding_name, (unsigned char *) UTF_8_NAME);
        utf_set_encoding();
        utf_cntr = 3;
      }
      else if (utf_read > 1 && xstrncmp((unsigned char *) UTF16BE_BOM, utf_buf, 2) == 0)
      {
        flag_bom_found = TRUE;
        sw_encoding = UTF_16BE;
        xstrcpy(input_record.bom_name, (unsigned char *) UTF_16BE_NAME);
        /* Maybe user wants to keep the short name */
        if (xstrcmp(output_record.encoding_name, (unsigned char *) UTF_16_NAME))
          xstrcpy(output_record.encoding_name, (unsigned char *) UTF_16BE_NAME);
        utf_set_encoding();
        utf_cntr = 2;
      }
      else if (utf_read > 1 && xstrncmp((unsigned char *) UTF16LE_BOM, utf_buf, 2) == 0)
      {
        /* check for UTF-32LE */
        if (utf_read > 3
            && utf_buf[2] == 0x00
            && utf_buf[3] == 0x00)
        {
          flag_bom_found = TRUE;
          sw_encoding = UTF_32LE;
          xstrcpy(input_record.bom_name, (unsigned char *) UTF_32LE_NAME);
          /* Maybe user wants to keep the short name */
          if (xstrcmp(output_record.encoding_name, (unsigned char *) UTF_32_NAME))
            xstrcpy(output_record.encoding_name, (unsigned char *) UTF_32LE_NAME);
          utf_set_encoding();
          utf_cntr = 4;
        }
        else
        {
          flag_bom_found = TRUE;
          sw_encoding = UTF_16LE;
          xstrcpy(input_record.bom_name, (unsigned char *) UTF_16LE_NAME);
          /* Maybe user wants to keep the short name */
          if (xstrcmp(output_record.encoding_name, (unsigned char *) UTF_16_NAME))
            xstrcpy(output_record.encoding_name, (unsigned char *) UTF_16LE_NAME);
          utf_set_encoding();
          utf_cntr = 2;
        }
      }
      else if (utf_read > 3 
          && utf_buf[0] == 0x00
          && utf_buf[1] == 0x00
          && utf_buf[2] == 0xfe
          && utf_buf[3] == 0xff)
      {
        flag_bom_found = TRUE;
        sw_encoding = UTF_32BE;
        xstrcpy(input_record.bom_name, (unsigned char *) UTF_32BE_NAME);
        /* Maybe user wants to keep the short name */
        if (xstrcmp(output_record.encoding_name, (unsigned char *) UTF_32_NAME))
          xstrcpy(output_record.encoding_name, (unsigned char *) UTF_32BE_NAME);
        utf_set_encoding();
        utf_cntr = 4;
      }

      /* update MD5 of preloaded buffer */
      if (sw_summary && sw_digest && fd == STDIN_FD)
      {
        MD5Update(&md5_in_context, utf_buf, utf_read);
        SHA512Input(&sha512_in_context, utf_buf, utf_read);
      }
    }
  }
  if (sw_debug)
    fprintf(stderr, "DEBUG: BOM found in input stream: %s.%s", input_record.bom_name, newline);

  return(flag_bom_found);
}


void utf_print_bom(fd)
int fd;
{
  /* Print out the BOM sequence 
   * whilst encoding sensitive */
  if (output_record.encoding == UTF_8)
    utf_print(fd, (unsigned char *) UTF8_BOM, 3);
  else if (output_record.encoding == UTF_16BE)
    utf_print(fd, (unsigned char *) UTF16BE_BOM, 2);
  else if (output_record.encoding == UTF_16LE)
    utf_print(fd, (unsigned char *) UTF16LE_BOM, 2);
  else if (output_record.encoding == UTF_32BE)
    utf_print(fd, (unsigned char *) UTF32BE_BOM, 4);
  else if (output_record.encoding == UTF_32LE)
    utf_print(fd, (unsigned char *) UTF32LE_BOM, 4);
}
