/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "validate_args.h"

/* function templates */
int  process_arg_regex(unsigned char *opt_s, unsigned char *s);


int process_arg_regex(opt_s, s)
unsigned char *opt_s;
unsigned char *s;
{
  /* translate regex arg to column structure */
  unsigned char *opt_delimiter;
  int i, j, k;
  int s_start, s_length, o_length, p_start, p_length, n_start;
  int cflags, eflags;
  int ret;

  /* initialise */
  ret           = TRUE;
  opt_delimiter = NULL;
  s_start       = 0;
  s_length      = xstrlen(s);
  cflags        = REG_EXTENDED; /* default */
  eflags        = 0;

  /* identify any quotes */
  if (s_length && (s[0] == '\'' || s[0] == '"'))
  {
    if (s_length && s[s_length - 1] == s[0])
      s_length--;
    s_start = 1;
    s_length--;
  }

  if (s_length == 0)
    ret = -21; /* invalid option "" - unknown. */

  p_start  = 0;
  p_length = 0;
  o_length = 0;
  /* determine the delimiter character used */
  i = s_start;
  while (i < s_length && s[i] != '/' && s[i] != '%')
    i++;
  if (i < s_length + 1)
  {
    opt_delimiter = &s[i];
    p_start = i + 1;
    o_length = i - s_start;
  }
  else
    ret = -22; /* invalid option - no option delimiter found. */

  /* process ../<pattern>/.. arg component */
  n_start = s_length;
  for (p_length = s_length - 1 ; p_length > s_start && s[p_length] != *opt_delimiter ; p_length--)
    n_start--;
  p_length = p_length - p_start;
  if (p_length < 1)
    ret = -23; /* invalid option - no pattern found. */
  else if (s[p_start + p_length] != *opt_delimiter)
    ret = -22; /* invalid option - option delimiter mismatched. */

  if (ret == TRUE)
  {
    /* process <cflags>/../.. arg component */
    for (i = s_start ; i < o_length ; i++)
    {
      switch(s[i])
      {
        /* regcomp() flags */
        case 'E' : cflags |= REG_EXTENDED;
                   cflags ^= REG_BASIC;
                   break;
  
        case 'B' : cflags |= REG_BASIC;
                   cflags ^= REG_EXTENDED;
                   break;
  
        case 'i' : cflags |= REG_ICASE;
                   break;
  
        case 'l' : cflags |= REG_LITERAL;
                   break;
  
        case 'U' : cflags |= REG_UNGREEDY; /* TRE uses uppercase U */
                   break;
  
        case 'r' : cflags |= REG_RIGHT_ASSOC;
                   break;
  
        case 'n' : cflags |= REG_NEWLINE;
                   break;
  
        /* regexec() flags */
        case 'b' : eflags |= REG_NOTBOL;
                   break;
  
        case 'e' : eflags |= REG_NOTEOL;
                   break;
  
        default  : ret = -24; /* invalid flags */
      }
    }
  }

  if (ret == TRUE)
  {
    /* allocate memory for regex structure */
    if (ncolumn)
      column = (struct_column *) realloc(column, ((ncolumn + 1) * sizeof(struct_column)));
    else
      column = (struct_column *) malloc(sizeof(struct_column));
    if (column == NULL) out_of_mem();

    /* set the type of column */
    column[ncolumn].type = OPT_REGEX;

    column[ncolumn].eflags = eflags;
    column[ncolumn].pattern_length = p_length;
    column[ncolumn].pattern = (unsigned char *) malloc((column[ncolumn].pattern_length + 1) * sizeof(unsigned char));
    if (column[ncolumn].pattern == NULL) out_of_mem();
    for (i = 0 ; i < column[ncolumn].pattern_length ; i++)
      column[ncolumn].pattern[i] = s[p_start + i];
    column[ncolumn].pattern[i] = '\0';

    /* process ../../<name> arg component */
    if (n_start < s_length)
    {
      /* find out how many tag namea have been specified */
      column[ncolumn].ntagname = 1;
      for (i = n_start ; i < s_length ; i++)
        if (s[i] == ',')
          column[ncolumn].ntagname++;

      /* allocate tag name memory */
      column[ncolumn].tagname = (unsigned char **) malloc(column[ncolumn].ntagname * sizeof(unsigned char *));
      if (column[ncolumn].tagname == NULL) out_of_mem();
      column[ncolumn].name = (unsigned char *) malloc((s_length - n_start + 1) * sizeof(unsigned char));
      if (column[ncolumn].name == NULL) out_of_mem();

      /* populate tag name structure */
      j = 0;
      k = 0;
      column[ncolumn].tagname[k] = &column[ncolumn].name[k];
      for (i = n_start ; i < s_length ; i++)
      {
        column[ncolumn].name[j] = s[i];
        if (s[i] == ',')
        {
          column[ncolumn].name[j] = '\0';
          k++;
          column[ncolumn].tagname[k] = &column[ncolumn].name[j + 1];
        }
        j++;
      }
      column[ncolumn].name[j] = '\0'; /* we added 1 to the original array size */
      /* Now trim leading/trailing whitespace */
      for (k = 0 ; k < column[ncolumn].ntagname ; k++)
      {
        /* remove trailing whitespace */
        for (j = xstrlen(column[ncolumn].tagname[k]) - 1 ; j >= 0 && isspace(column[ncolumn].tagname[k][j]) ; j--)
          column[ncolumn].tagname[k][j] = '\0';
        /* ignore leading whitespace */
        for (j = 0 ; j < xstrlen(column[ncolumn].tagname[k]) && isspace(column[ncolumn].tagname[k][j]) ; j++)
          ;
        if (j > 0 && j < xstrlen(column[ncolumn].tagname[k]))
          column[ncolumn].tagname[k] = &column[ncolumn].tagname[k][j];
      }

      /* get rid of any empty names */
      i = 0;
      j = column[ncolumn].ntagname;
      while (i < j)
      {
        if (xstrlen(column[ncolumn].tagname[i]) == 0)
        {
          for (k = i ; k + 1 < j ; k++)
            column[ncolumn].tagname[k] = column[ncolumn].tagname[k + 1];
          column[ncolumn].ntagname--;
        }
        i++;
      }
    }
    else
      column[ncolumn].ntagname = 0;

    /* if scope exists process it */
    s_length = xstrlen(opt_s);
    for (s_start = 0 ; s_start < s_length && opt_s[s_start] != ':' ; s_start++ )
      ;
    if (s_start + 1 < s_length)
    {
      s_start++;
      s_length -= s_start;
      /* add scope to the regex_opt structure */
      if (populate_scope(OPT_REGEX, -1, &opt_s[s_start], s_length))
      {
        column[ncolumn].nscope = nscope;
        column[ncolumn].scope  = scope;
      }
      else
        ret = -15; /* invalid <scope> provided */
    }
    else
    {
      column[ncolumn].nscope = 0;
      column[ncolumn].scope  = NULL;
    }
    column[ncolumn].scope_seq = 0;
  }

  /* compile the RE pattern */
  if (ret == TRUE)
  {
    column[ncolumn].pmatch = NULL;
    if (regncomp(&column[ncolumn].pcomp, (char *) column[ncolumn].pattern, column[ncolumn].pattern_length, cflags) == 0)
    {
      if (column[ncolumn].pcomp.re_nsub > 0)
        column[ncolumn].pmatch = (regmatch_t *) malloc((column[ncolumn].pcomp.re_nsub + 1) * sizeof(regmatch_t));
      else
        column[ncolumn].pmatch = (regmatch_t *) malloc(sizeof(regmatch_t));
      if (column[ncolumn].pmatch == NULL) out_of_mem();
    }
    else
      ret = -25; /* invalid RE pattern */

    ncolumn++;
  }

  return(ret);
}
