/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "validate_args.h"

/* function templates */
int process_arg_scoped(int opt_type, unsigned char *s);
int populate_scope(int opt_type, int level, unsigned char *scope_opt, int s_length);


int process_arg_scoped(opt_type, s)
int opt_type;
unsigned char *s;
{
  /* translate arg option to relevant parts */
  int s_length, s_start, scope_start;
  int i, j, k;
  char value[10];
  int level;

  /* initialise */
  level               = 0;
  delimiter_byte_mode = FALSE;
  nscope              = 0;
  scope               = NULL;

  /* find the start of the interesting part of the arg option */
  s_start = 0;
  s_length = xstrlen(s);
  while (s_start < s_length && !isdigit(s[s_start]) && s[s_start] != ':')
    s_start++;

  /* check if a scope exists in the option */
  scope_start = 0;
  for (i = s_start ; i < s_length ; i++)
  {
    if (s[i] == ':')
    {
      scope_start = i + 1;
      break;
    }
  }

  /* if no scope option then limited processing required therefore return */
  if (s_start == 0)                    /* invalid option "" - unknown */
    return(-11);
  else if (s_start == scope_start - 1) /* invalid option e.g. "-F:<scope>" - no <level> provided */
    return(-12);
  else if (scope_start == s_length)    /* invalid option e.g. "-F<level>:" - no <scope> provided */
    return(-13);
  else if (s_start == s_length)    /* This condition should already be handled by the calling function - e.g. plain option "-F" */
    return(2);

  j = 0;
  k = 0;

  /* get the level, and if delimiter scope also get byte mode switch */
  if (scope_start == 0)
  {
    if (opt_type == OPT_DELIMITER && s_length > 0 && s[s_length - 1] == 'b')
    {
      delimiter_byte_mode = TRUE;
      k = s_length - 1;
    }
    else
      k = s_length;
  }
  else
  {
    if (opt_type == OPT_DELIMITER && scope_start - 2 > 0 && s[scope_start - 2] == 'b')
    {
      delimiter_byte_mode = TRUE;
      k = scope_start - 2;
    }
    else
      k = scope_start - 1;
  }

  for (i = s_start ; i < k ; i++)
    value[j++] = s[i];
  value[j++] = '\0';
  level = atoi(value);
  /* atoi() is too smart */
  for (i = 0 ; i < j - 1 ; i++)
    if (!isdigit((int) value[i]))
      level = -1;

  /* if no scope then return */
  if (scope_start == 0)
    return(level);

  /* set master scoping switch to on  - mainly for effeciency elsewhere in xmlfy */
  if (opt_type == OPT_DELIMITER)
    sw_delimiter_scoping = TRUE;

  /* get the scope */
  s_length = xstrlen(&s[scope_start]);
  /* Calling function e.g. populate_delimiter() will pick up this scope structure */
  if (populate_scope(opt_type, level, &s[scope_start], s_length))
    return(level);
  else
    return(-15); /* invalid <scope> provided */
}


int populate_scope(opt_type, level, scope_opt, s_length)
int opt_type;
int level;
unsigned char *scope_opt;
int s_length;
{
  /* populate the scope struct from the supplied argument */
  int i, j;
  char value[10];
  int i_value;
  int flag_valid_scope;

  /* initialise */
  nscope           = 0;
  scope            = NULL;
  flag_valid_scope = TRUE;

  /* ensure legal scope characters */
  for (i = 0 ; i < s_length ; i++)
  {
    switch (scope_opt[i])
    {
      case '0' : 
      case '1' : 
      case '2' : 
      case '3' : 
      case '4' : 
      case '5' : 
      case '6' : 
      case '7' : 
      case '8' : 
      case '9' : 
      case '$' : 
      case '-' : 
      case ',' : 
      case 'r' : 
                 break;

      default :  flag_valid_scope = FALSE;
    }
  }

  if (flag_valid_scope)
  {
    /* calculate number of scope structures required */
    nscope = 1;
    for (i = 0 ; i < s_length ; i++)
    {
      if (scope_opt[i] == ',')
        nscope++;
    }

    /* allocate scopes */
    scope = (struct_scope *) malloc(nscope * sizeof(struct_scope));
    if (scope == NULL) out_of_mem();
  
    /* initialise scope structs */
    for (i = 0 ; i < nscope ; i++)
    {
      scope[i].min     = RANGE_UNSET;
      scope[i].max     = RANGE_UNSET;
      scope[i].restart = FALSE;
    }

    /* populate scope structure */
    j           = 0;
    i_value     = 0;
    value[0]    = '\0';
    for (i = 0 ; i < s_length ; i++)
    {
      switch (scope_opt[i])
      {
        case '$' : if (scope[j].min == RANGE_UNSET)
                       flag_valid_scope = FALSE; /* in this release the last record must be in a range */
                     else
                       scope[j].max = -1;
                     i_value = 0;
                     break;

        case '-' : if (i_value)
                   {
                     if (scope[j].min == RANGE_UNSET)
                     {
                       scope[j].min = atol(value) - 1;
                       if (scope[j].min < 0)
                         flag_valid_scope = FALSE;
                     }
                   }
                   else
                     scope[j].min = 0;
                   i_value = 0;
                   /* pre-populate max field in case not supplied */
                   scope[j].max = -1;
                   break;

        case 'r' : scope[j].restart = TRUE;
                   if (opt_type == OPT_DELIMITER)
                   {
                     if (restarting_delimiter_scope_level < 0)
                       restarting_delimiter_scope_level = level;
                     else if (level < restarting_delimiter_scope_level)
                       restarting_delimiter_scope_level = level;
                   }
                   break;

        case ',' : if (i_value)
                   {
                     if (scope[j].min == RANGE_UNSET)
                     {
                       scope[j].min = atol(value) - 1;
                       if (scope[j].min < 0)
                         flag_valid_scope = FALSE;
                       /* pre-populate max field in case not supplied */
                       scope[j].max = scope[j].min;
                     }
                     else
                     {
                       scope[j].max = atol(value) - 1;
                       if (scope[j].max < 0)
                         flag_valid_scope = FALSE;
                     }
                   }
                   i_value = 0;
                   /* quick sanity check */
                   if (scope[j].max > 0 && scope[j].min > scope[j].max)
                     flag_valid_scope = FALSE;
                   j++;
                   break;

        default :  value[i_value++] = scope_opt[i];
                   value[i_value]   = '\0';
      }
    }
    if (i_value)
    {
      if (scope[j].min == RANGE_UNSET)
      {
        scope[j].min = atol(value) - 1;
        if (scope[j].min < 0)
          flag_valid_scope = FALSE;
        /* pre-populate max field in case not supplied */
        scope[j].max = scope[j].min;
      }
      else
      {
        scope[j].max = atol(value) - 1;
        if (scope[j].max < 0)
          flag_valid_scope = FALSE;
      }
      /* quick sanity check */
      if (scope[j].max > 0 && scope[j].min > scope[j].max)
        flag_valid_scope = FALSE;
    }
    /* quick sanity check */
    for (i = 0 ; i < nscope ; i++)
      if (scope[i].min == RANGE_UNSET && scope[i].max == RANGE_UNSET)
        flag_valid_scope = FALSE;
  }
  
  /* scope processing now complete.
   * Calling function e.g. populate_delimiter() will pick up this scope structure */
  return(flag_valid_scope);
}
