/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlfy.h"

const char progname[] = PROGNAME;

struct_schema_parent  schema_parent;
struct_schema_element *schema_element;
struct_schema_record  *schema_record;

unsigned char *schema_filename;
unsigned char *schema_system_identifier = NULL;
unsigned char *schema_xsd_target_ns     = NULL;
unsigned char *schema_parent_name;
unsigned char *schema_line_name;
unsigned char *schema_linenumber;

unsigned char *default_schema_parent_name;
unsigned char *default_schema_line_name;
unsigned char *default_schema_field_name;

/* level handling */
int nlevels = DEFAULT_NLEVELS;

/* switches */
int sw_xml_format            = SW_XML;
int sw_encoding              = UTF_8;
int sw_schema                = SCHEMA_DEFAULT;
int sw_printonly_header      = TRUE;
int sw_printonly_footer      = TRUE;
int sw_printonly_rtagopen    = TRUE;
int sw_printonly_rtagclose   = TRUE;
int sw_printonly_records     = TRUE;
int sw_printonly_first_opt   = TRUE;
int sw_char_escape_amp       = TRUE;
int sw_char_escape_lt        = TRUE;
int sw_char_escape_gt        = TRUE;
int sw_char_escape_quot      = TRUE;
int sw_char_escape_apos      = TRUE;
int sw_char_escape_brvbar    = TRUE;
int sw_process_output_digest = TRUE;
int sw_digest                = TRUE;
int sw_line_tags             = FALSE;
int sw_field_quoting         = FALSE;
int sw_field_quoting_span    = FALSE;
int sw_field_expel_empty     = FALSE;
int sw_field_empty_valid     = FALSE;
int sw_field_trim            = FALSE;
int sw_blank_lines           = FALSE;
int sw_key_value_pairs       = FALSE;
int sw_kvp_levels_all        = FALSE;
int sw_bom                   = FALSE;
int sw_ascii_to_utf          = FALSE;
int sw_delimiter_scoping     = FALSE;
int sw_trimtagclose          = FALSE;
int sw_summary               = FALSE;
int sw_debug                 = FALSE;
struct_sw_print **sw_print, sw_print_default;

/* key/value pair handling */
struct_kvp_tag **kvp_tag;

/* input line handling */
struct_input_record input_record;
unsigned char utf_buf[BUFSIZE];
int utf_flag_EOF;
int utf_read;
int utf_cntr;

/* delimiter handling */
struct_delimiter **delimiter;
unsigned char delimiter_newline[2];
unsigned char delimiter_space[2];
unsigned char delimiter_tab[2];
unsigned char delimiter_cr[2];
int delimiter_byte_mode;

/* expel records/fields handling */
struct_expel *expel;
int nexpel; 

/* column range/name handling */
struct_column *column;
int ncolumn; 
int flag_input_has_composites;

/* custom element attribute handling */
int ninsert_attr;
struct_insert_attr *insert_attr;

/* custom element tag handling */
int ninsert_tag;
struct_insert_tag *insert_tag;

/* delimiter scope handling */
struct_scope *scope;
int nscope;
int restarting_delimiter_scope_level = -1;

/* output line handling */
struct_output_record output_record;

/* output newline handling */
unsigned char newline[3];

/* utf format handling function pointers */
fptr_void strcpy_output_record;
fptr_void strncpy_output_record;
fptr_void strcat_output_record;
fptr_void strncat_output_record;

/* memory block growth handling triggers */
/* factor MEMSEG_THRESHOLD for unforseen segment overruns */
long memseg_nfields       = DEFAULT_NFIELDS - MEMSEG_THRESHOLD;
long memseg_input_record  = LINESIZE - MEMSEG_THRESHOLD;
long memseg_output_record = LINESIZE - MEMSEG_THRESHOLD;
/* memory allocation function call counters */
int  memadd_nfields_cntr       = 0;
int  memadd_input_record_cntr  = 0;
int  memadd_output_record_cntr = 0;

/* timestamp handling */
time_t start_ts, ts;

/* schema match start level handling */
int schema_match_start_level = 1; /* default legacy mode */
unsigned char *schema_match_start_name = NULL;

/* schema file option handling */
unsigned char **schema_argv;
int schema_argc;

/* quotechar handling */
int quotechar_type = 0;
unsigned char *quotechars;
int quotechars_length;
int i_quotechar;

/* XML/SOAP version implementation handling */
int version_xml  = XML_1_0;
int version_soap = SOAP_1_1;

/* summary handling */
int summary_fd = STDOUT_FD;

/* MD5 signature */
MD5_CTX md5_in_context;
MD5_CTX md5_out_context;
/* SHA512 signature */
SHA512Context sha512_in_context;
SHA512Context sha512_out_context;

/* HTML output handling */
unsigned char *html_parent_tagname;
unsigned char *html_child_tagname;

/* error status handling */
int err_status = 0;

/* function templates */
void allocate_mem();
void cleanup();
void out_of_mem();
/* xmlfy simplified libc like routines to handle unsigned char */
void xstrcpy(unsigned char *s_dst, const unsigned char *s_src);
void xstrncpy(unsigned char *s_dst, const unsigned char *s_src, long n);
int  xstrcmp(const unsigned char *s1, const unsigned char *s2);
int  xstrncmp(const unsigned char *s1, const unsigned char *s2, long n);
int  xstrlen(const unsigned char *s1);
#ifdef XMLFY_WIN32
void set_win32_binary_mode();
#endif /* XMLFY_WIN32 */


int main(argc, argv) 
int argc;
unsigned char **argv;
{
  /* show starts here */
  xstrcpy(newline, (unsigned char *) NEWLINE_UNIX);
#ifdef XMLFY_WIN32
  set_win32_binary_mode();
  xstrcpy(newline, (unsigned char *) NEWLINE_DOS);
#endif /* XMLFY_WIN32 */
  time(&start_ts);
  allocate_mem();
  if (err_status == 0) validate_args(argc, argv, ARGS_CMD);
  if (sw_xml_format == SW_XML || sw_xml_format == SW_SOAP || sw_xml_format == SW_HTML)
  {
    /* add XML */
    if (err_status == 0) schema();
    if (err_status == 0) xmlout(argc, argv);
  }
  else if (sw_xml_format == SW_NOXML)
  {
    /* convert input stream 
     * E.g. Convert reserved XML chars to escaped sequences
       and/or convert input to wide UTF encoding */
    if (err_status == 0) noxml();
  }
  else if (sw_xml_format == SW_UNXML)
  {
    /* remove XML */
    if (err_status == 0) unxml();
  }
  cleanup();

  return (err_status);
}


void allocate_mem()
{
  /* Allocate initial memory */
  int i;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: allocate_mem%s", newline);

  default_schema_parent_name = (unsigned char *) malloc(sizeof(DEFAULT_STR_ELEMENT_PARENT) + sizeof(unsigned char));
  if (default_schema_parent_name == NULL) out_of_mem();
  xstrcpy(default_schema_parent_name, (unsigned char *) DEFAULT_STR_ELEMENT_PARENT);

  default_schema_line_name = (unsigned char *) malloc(sizeof(DEFAULT_STR_ELEMENT_LINE) + sizeof(unsigned char));
  if (default_schema_line_name == NULL) out_of_mem();
  xstrcpy(default_schema_line_name, (unsigned char *) DEFAULT_STR_ELEMENT_LINE);

  default_schema_field_name = (unsigned char *) malloc(sizeof(DEFAULT_STR_ELEMENT_FIELD) + sizeof(unsigned char)); 
  if (default_schema_field_name == NULL) out_of_mem();
  xstrcpy(default_schema_field_name, (unsigned char *) DEFAULT_STR_ELEMENT_FIELD);

  schema_linenumber = (unsigned char *) malloc(sizeof(DEFAULT_STR_ELEMENT_LINENUMBER) + sizeof(unsigned char));
  if (schema_linenumber == NULL) out_of_mem();
  xstrcpy(schema_linenumber, (unsigned char *) DEFAULT_STR_ELEMENT_LINENUMBER);

  input_record.line = (unsigned char *) malloc(LINESIZE * sizeof(unsigned char));
  if (input_record.line == NULL) out_of_mem();

  input_record.field = (struct_input_field **) malloc(DEFAULT_NFIELDS * sizeof(struct_input_field *));
  if (input_record.field == NULL) out_of_mem();
  for (i = 0 ; i < DEFAULT_NFIELDS ; i++)
  {
    input_record.field[i] = (struct_input_field *) malloc(sizeof(struct_input_field));
    if (input_record.field[i] == NULL) out_of_mem();
  }

  output_record.line = (unsigned char *) malloc(LINESIZE * sizeof(unsigned char));
  if (output_record.line == NULL) out_of_mem();

  delimiter = (struct_delimiter **) malloc(DEFAULT_NLEVELS * sizeof(struct_delimiter *));
  if (delimiter == NULL) out_of_mem();
  for (i = 0 ; i < DEFAULT_NLEVELS ; i++)
  {
    delimiter[i] = (struct_delimiter *) malloc(sizeof(struct_delimiter));
    if (delimiter[i] == NULL) out_of_mem();
  }

  sw_print = (struct_sw_print **) malloc(DEFAULT_NLEVELS * sizeof(struct_sw_print *));
  if (sw_print == NULL) out_of_mem();
  kvp_tag = (struct_kvp_tag **) malloc(DEFAULT_NLEVELS * sizeof(struct_kvp_tag *));
  if (kvp_tag == NULL) out_of_mem();
  for (i = 0 ; i < DEFAULT_NLEVELS ; i++)
  {
    sw_print[i] = (struct_sw_print *) malloc(sizeof(struct_sw_print));
    if (sw_print[i] == NULL) out_of_mem();
    kvp_tag[i] = (struct_kvp_tag *) malloc(sizeof(struct_kvp_tag));
    if (kvp_tag[i] == NULL) out_of_mem();
  }

  output_record.field_cntr = (int *) malloc(DEFAULT_NLEVELS * sizeof(int));
  if (output_record.field_cntr == NULL) out_of_mem();

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: allocate_mem%s", newline);
}


void memadd_nfields()
{
  /* re-allocate more memory for more fields */
  long i, multiplier;

  /* To reduce the frequency of calling this function increase memory allocation 
   * the more often this function is called. */
  memadd_nfields_cntr++;
  if (memadd_nfields_cntr < 10)
    multiplier = 1;
  else if (memadd_nfields_cntr < 20)
    multiplier = 10;
  else if (memadd_nfields_cntr < 30)
    multiplier = 100;
  else
    multiplier = 500;
  
  i = memseg_nfields + MEMSEG_THRESHOLD;
  memseg_nfields = ((input_record.nfields + MEMSEG_THRESHOLD) / DEFAULT_NFIELDS + 1) * DEFAULT_NFIELDS * multiplier - MEMSEG_THRESHOLD;
  input_record.field = (struct_input_field **) realloc(input_record.field, (memseg_nfields + MEMSEG_THRESHOLD) * sizeof(struct_input_field *));
  if (input_record.field == NULL) out_of_mem();
  while (i < memseg_nfields + MEMSEG_THRESHOLD)
  { 
    input_record.field[i] = (struct_input_field *) malloc(sizeof(struct_input_field));
    if (input_record.field[i] == NULL) out_of_mem();
    
    if (sw_schema != SCHEMA_DEFAULT)
    {
      input_record.field[i]->schema_xpath = (int *) malloc(nlevels * sizeof(int));
      if (input_record.field[i]->schema_xpath == NULL) out_of_mem();
    }
    i++;
  }
}


void memadd_input_record()
{
  int multiplier;

  /* To reduce the frequency of calling this function increase memory allocation 
   * the more often this function is called. */
  memadd_input_record_cntr++;
  if (memadd_input_record_cntr < 10)
    multiplier = 1;
  else if (memadd_input_record_cntr < 20)
    multiplier = 10;
  else if (memadd_input_record_cntr < 30)
    multiplier = 100;
  else
    multiplier = 500;

  /* re-allocate more memory for more input_record */
  memseg_input_record = ((input_record.length + MEMSEG_THRESHOLD) / LINESIZE + 1) * LINESIZE * multiplier - MEMSEG_THRESHOLD;
  input_record.line = (unsigned char *) realloc(input_record.line, (memseg_input_record + MEMSEG_THRESHOLD) * sizeof(unsigned char));
  if (input_record.line == NULL) out_of_mem();
}


void memadd_output_record(msize)
long msize;
{
  int multiplier;

  /* To reduce the frequency of calling this function increase memory
   * allocation the more often this function is called. */
  memadd_output_record_cntr++;
  if (memadd_output_record_cntr < 10)
    multiplier = 1;
  else if (memadd_output_record_cntr < 20)
    multiplier = 10;
  else if (memadd_output_record_cntr < 30)
    multiplier = 100;
  else
    multiplier = 500;

  /* re-allocate more memory for more output_record line */
  memseg_output_record = ((msize + MEMSEG_THRESHOLD) / LINESIZE + 1) * LINESIZE * multiplier - MEMSEG_THRESHOLD;
  output_record.line = (unsigned char *) realloc(output_record.line, (memseg_output_record + MEMSEG_THRESHOLD) * sizeof(unsigned char));
  if (output_record.line == NULL) out_of_mem();
}


void cleanup()
{
  /* Clean up before closure */
  int i, j;

  if (sw_debug)
    fprintf(stderr, "DEBUG: start function: cleanup%s", newline);
  
  /* 
   * close all files 
   */

  if (sw_summary == SUMMARY_FILE && summary_fd > 0)
    (void) close(summary_fd);

  /* 
   * free all allocated memory 
   */

  /* free records and their fields */
  for (i = 0 ; i < schema_parent.nrecords; i++)
  {
    for (j = 0 ; j < schema_record[i].nfields; j++)
    {
      (void) free(schema_record[i].field[j].name);
    }
    (void) free(schema_record[i].field);
  }
  (void) free(schema_record);

  /* free elements */
  for (i = 0 ; i < schema_parent.nelements; i++)
  {
    (void) free(schema_element[i].record);
    if (schema_element[i].element_helper != NULL)
    {
      for (j = 0 ; j < schema_element[i].element_helper->nxpaths; j++)
        (void) free(schema_element[i].element_helper->match_helper[j].xpath);
      (void) free(schema_element[i].element_helper->match_helper);
      (void) free(schema_element[i].element_helper);
    }
  }
  (void) free(schema_element);

  (void) free(default_schema_parent_name);
  (void) free(default_schema_line_name);
  (void) free(default_schema_field_name);
  /* (void) free(schema_parent_name); = reference pointer to other pointer */
  (void) free(schema_linenumber);

  /* free input/output storage */
  (void) free(input_record.line);
  for (i = 0 ; i < memseg_nfields + MEMSEG_THRESHOLD ; i++)
  {
    if (sw_schema != SCHEMA_DEFAULT)
    {
      (void) free(input_record.field[i]->schema_xpath); 
    }
    (void) free(input_record.field[i]); 
  }
  (void) free(input_record.field); 
  (void) free(output_record.line);

  /* free level objects */
  for (i = 0 ; i < nlevels ; i++)
  {
    if (delimiter[i]->ndelimiters)
    {
      for (j = 0 ; j < delimiter[i]->ndelimiters ; j++)
      {
        (void) free(delimiter[i]->utf_d[j]);
        if(delimiter[i]->nscope[j])
          (void) free(delimiter[i]->scope[j]);
      }
      (void) free(delimiter[i]->utf_d);
      (void) free(delimiter[i]->utf_length);
      (void) free(delimiter[i]->arg_d);
      (void) free(delimiter[i]->arg_length);
      (void) free(delimiter[i]->byte_mode);
      (void) free(delimiter[i]->nscope);
      (void) free(delimiter[i]->scope);
      (void) free(delimiter[i]->scope_seq);
    }
    (void) free (delimiter[i]);
    (void) free (sw_print[i]);
    (void) free (kvp_tag[i]);
  }
  (void) free(delimiter); 
  (void) free(sw_print); 
  (void) free(kvp_tag); 
  (void) free(output_record.field_cntr); 

  /* free schema args storage */
  if (schema_argc)
  {
    for (i = 1 ; i < schema_argc ; i++)
      (void) free(schema_argv[i]);

    (void) free(schema_argv);
  }

  /* free expel storage */
  if (nexpel)
  {
    for (i = 0 ; i < nexpel ; i++)
    {
      if (expel[i].n_erecords)
        (void) free(expel[i].erecords);
      if (expel[i].n_efields)
        (void) free(expel[i].efields);
    }

    (void) free(expel);
  }

  /* free column storage */
  if (ncolumn)
  {
    for (i = 0 ; i < ncolumn ; i++)
    {
      if (column[i].nscope)
        (void) free(column[i].scope);
      if (column[i].type == OPT_REGEX)
      {
        (void) free(column[i].pattern);
        if (column[i].pmatch != NULL)
          (void) free(column[i].pmatch);
        (void) regfree(&column[i].pcomp);
        if (column[i].ntagname)
        {
          (void) free(column[i].name);
          (void) free(column[i].tagname);
        }
      }
    }
    (void) free(column);
  }

  /* free custom attribute storage */
  if (ninsert_attr)
  {
    for (i = 0 ; i < ninsert_attr ; i++)
      if (insert_attr[i].nscope)
        (void) free(insert_attr[i].scope);

    (void) free(insert_attr);
  }

  /* free custom tag storage */
  if (ninsert_tag)
  {
    for (i = 0 ; i < ninsert_tag ; i++)
    {
      if (insert_tag[i].type == TAG_INSERT_FILE || insert_tag[i].type == TAG_INSERT_XMLFILE)
        (void) free(insert_tag[i].value);
      if (insert_tag[i].nscope)
        (void) free(insert_tag[i].scope);
    }

    (void) free(insert_tag);
  }

  /* free XSD targetNamespace */
  if (schema_xsd_target_ns != NULL)
    (void) free(schema_xsd_target_ns);

  /* free HTML tagnames */
  if (sw_xml_format == SW_HTML)
  {
    (void) free(html_parent_tagname);
    (void) free(html_child_tagname);
  }

  if (sw_debug)
    fprintf(stderr, "DEBUG: end function: cleanup%s", newline);
}


void out_of_mem()
{
  /* No memory, ladies and gentlemen the show is over.*/

  err_status = -10;
  fprintf(stderr, "%s: ERROR: Cannot allocate memory (malloc/realloc): %d.%s", progname, err_status, newline);
  exit(err_status);
}


/* =============================================================
 * xmlfy simplified libc like routines to handle unsigned char.
 * NOTE: We still use standard libc routines depending on need.
 * =============================================================
 */

void xstrcpy(unsigned char *s_dst, const unsigned char *s_src)
{
  /* Copy chars from s_src to s_dst until \0 is reached */
  long i = 0;

  do {
    s_dst[i] = s_src[i];
  } while (s_src[i++] != '\0');
}

void xstrncpy(unsigned char *s_dst, const unsigned char *s_src, long n)
{
  /* Copy n chars from s_src to s_dst and pad s_dst if s_src < n */
  long i = 0;

  while (i < n && s_src[i] != '\0');
  {
    s_dst[i] = s_src[i];
    i++;
  }

  /* if s_src length < n then pad s_dst with \0 */
  while (i < n)
    s_dst[i++] = '\0';
}

int xstrcmp(const unsigned char *s1, const unsigned char *s2)
{
  /* Check if s1 == s2 and ONLY return 0 (match) or 1 (nomatch) */
  int ret = 1;
  long i = 0;

  while (s1[i] == s2[i])
  {
    if (s1[i] == '\0' || s2[i] == '\0')
    {
      if (s1[i] == s2[i])
        ret = 0;
      break;
    }
    i++;
  }

  return (ret);
}

int xstrncmp(const unsigned char *s1, const unsigned char *s2, long n)
{
  /* Check if s1 == s2 for up to n chars and ONLY return 0 (match) or 1 (nomatch) */
  int ret = 1;
  long i = 0;

  while (i < n && s1[i] == s2[i])
  {
    if (s1[i] == '\0' || s2[i] == '\0')
    {
      if (s1[i] == s2[i])
        ret = 0;
      break;
    }
    i++;
  }

  if (i == n)
    ret = 0;

  return (ret);
}

int xstrlen(const unsigned char *s1)
{
  /* Return the length of s1 */
  int i = 0;

  while (s1[i] != '\0')
    i++;

  return (i);
}


#ifdef XMLFY_WIN32
void set_win32_binary_mode()
{
  /* Windows machines open STDIN and STDOUT in text mode by default.
   * This means that "\n" characters are automatically translated to
   * "\r\n" characters whether you like it or not. This is fine for ASCII
   * and UTF-8 encoding, but corrupts the input and output for wide UTF
   * encoded streams.
   * I.e. the translation of "\n" should be "\000\r\000\n" for UTF-16BE
   * encoding, but that doesn't happen so we open STDIN and STDOUT
   * in binary mode and do it ourselves.
   * This Window's feature is not UTF but more like WTF!
   */

  /* This is a windows only function */
  (void) _setmode(STDIN_FD, _O_BINARY);
  (void) _setmode(STDOUT_FD, _O_BINARY);
  (void) _setmode(STDERR_FD, _O_BINARY);
}
#endif /* XMLFY_WIN32 */
