/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#ifndef XMLFY_H
#define XMLFY_H

/* header files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifndef XMLFY_WIN32
#include <unistd.h>
#endif /* XMLFY_WIN32 */
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "regex_tre/regex.h"
#include "digest/md5.h"
#include "digest/sha512.h"

/* saga control */
#define PROGNAME              "xmlfy"
#define VERSION               "1.5.6"

/* language control */
#define LANG_EN               /* English. Whose English? */

/* truth control */
#define TRUE                  1
#define FALSE                 0

/* IO control */
#define STDIN_FD              0
#define STDOUT_FD             1
#define STDERR_FD             2
#define BUFSIZE               512  /* block IO buffer size */
#define LINESIZE              2048 /* initial input/output record buffer size */
#define MEMSEG_THRESHOLD      10   /* watermark for triggering mem allocation */

/* XML control */
#define SW_XML                1
#define SW_SOAP               2
#define SW_HTML               3
#define SW_NOXML              8
#define SW_UNXML              9
#define XML_1_0               10
#define XML_1_1               11
#define SOAP_1_1              20
#define SOAP_1_2              21

/* UTF control */
#define UTF_8                 1
/* UTF_16 is not definite enough */
#define UTF_16BE              11
#define UTF_16LE              12
/* UTF_32 is not definite enough */
#define UTF_32BE              21
#define UTF_32LE              22
#define UTF_8_NAME            "UTF-8"
#define UTF_16_NAME           "UTF-16"
#define UTF_16BE_NAME         "UTF-16BE"
#define UTF_16LE_NAME         "UTF-16LE"
#define UTF_32_NAME           "UTF-32"
#define UTF_32BE_NAME         "UTF-32BE"
#define UTF_32LE_NAME         "UTF-32LE"
#define BOM_IGNORE_NAME       "ignore"
#define BOM_NONE_NAME         "none"
#define ENCODING_NAME_SIZE    10
#define UTF_8_INCREMENTER     1
#define UTF_16_INCREMENTER    2
#define UTF_32_INCREMENTER    4

/* default schema structure
 * in DTD
 *   <!ELEMENT xmlfy (line*)>
 *   <!ELEMENT line (field*)>
 *   <!ELEMENT field (#PCDATA)>
 * in RNC
 *   start = xmlfy 
 *   xmlfy = element xmlfy { line* }
 *   line  = element line  { field* }
 *   field = element field { text }
*/
#define DEFAULT_STR_ELEMENT_PARENT           "xmlfy"
#define DEFAULT_STR_ELEMENT_LINE             "line"
#define DEFAULT_STR_ELEMENT_FIELD            "field"
#define DEFAULT_STR_ELEMENT_LINENUMBER       "linenumber"
#define DEFAULT_STR_ATTRIBUTE_NUMBER         "number"
#define DEFAULT_STR_ATTRIBUTE_TIMESTAMP_DATE "timestamp_date"
#define DEFAULT_STR_ATTRIBUTE_TIMESTAMP_SECS "timestamp_secs"
#define DEFAULT_STR_ATTRIBUTE_LEVEL          "level"
#define DEFAULT_STR_ATTRIBUTE_DELIMITER      "delimiter"
#define DEFAULT_STR_TAG_LEVEL                "L"
#define DEFAULT_NFIELDS                      100 /* initial nfields */
#define DEFAULT_NLEVELS                      3   /* initial levels 0, 1, 2 */
/* SOAP */
#define DEFAULT_STR_SOAP_URI_1_1             "http://www.w3.org/2001/12"
#define DEFAULT_STR_SOAP_URI_1_2             "http://www.w3.org/2003/05"
#define DEFAULT_STR_SOAP_URI_XMLFY           "http://xmlfy.sourceforge.net/soap-xmlfy"
#define DEFAULT_STR_SOAP_URI_XMLFY_HEADER    "http://xmlfy.sourceforge.net/soap-xmlfyHeader"
#define DEFAULT_STR_SOAP_XS_ENVELOPE         "env"
#define DEFAULT_STR_SOAP_XS_XMLFY            "x"
#define DEFAULT_STR_SOAP_XS_XMLFY_HEADER     "xh"
/* XSD schema */
#define DEFAULT_STR_XSD_INSTANCE             "http://www.w3.org/2001/XMLSchema-instance"
/* HTML */
#define DEFAULT_STR_HTML5                    "<!DOCTYPE html>"
#define DEFAULT_STR_HTML_TABLE               "table"
#define DEFAULT_STR_HTML_TR                  "tr"
#define DEFAULT_STR_HTML_TD                  "td"
#define DEFAULT_STR_HTML_P                   "p"
#define DEFAULT_STR_HTML_UL                  "ul"
#define DEFAULT_STR_HTML_LI                  "li"

/* Byte-Order-Mark (BOM) control */
#define UTF8_BOM                             "\357\273\277"      /* U+FEFF */
#define UTF16BE_BOM                          "\376\377"          /* U+FEFF */
#define UTF16LE_BOM                          "\377\376"          /* U+FFFE */
#define UTF32BE_BOM                          "\000\000\376\377"  /* U+FEFF */
#define UTF32LE_BOM                          "\377\376\000\000"  /* U+FFFE */

/* match control */
#define MATCH_ONE                  0
#define MATCH_NONE_OR_ONE          1
#define MATCH_NONE_OR_MANY         2
#define MATCH_ONE_OR_MANY          3

/* field control */
#define FIELD_CDATA                0
#define FIELD_SUBRECORD            1
#define FIELD_KEYTAG_ELEMENT       10
#define FIELD_KEYTAG_RECORD        11
#define FIELD_EXPELLED             99

/* tag control */
#define TAG_START                  1
#define TAG_END                    2
#define TAG_EMPTY                  3
#define TAG_PROLOGUE               4
#define TAG_DOCTYPE                5

/* delimiter control */
#define NEWLINE                    '\n'
#define WHITESPACE                 ' '
#define QUOTECHAR                  '\"'
#define DELIMITER_INDEX_UNMATCHED  -1
#define DELIMITER_INDEX_WHITESPACE -2
#define DELIMITER_INDEX_NEWLINE    -3
#define DELIMITER_INDEX_COLUMN     -4
#define DELIMITER_INDEX_REGEX      -5

/* timestamp control */
#define TIMESTAMP_DATE             0
#define TIMESTAMP_SECS             1

/* option control */
#define ARGS_CMD                   0
#define ARGS_SCHEMA                1
#define DTD_ARGS_TOKEN             "<!-- xmlfy-args:"
#define DTD_ARGS_TOKEN_LENGTH      16
#define RNC_ARGS_TOKEN             "## xmlfy-args:"
#define RNC_ARGS_TOKEN_LENGTH      14

/* schema control */
#define SCHEMA_DEFAULT             0
#define SCHEMA_DTD                 1
#define SCHEMA_RNC                 2
#define SCHEMA_XSD                 4

/* expel control */
#define EXPEL_NONE                 0
#define EXPEL_RECORD               1
#define EXPEL_FIELDS               2

/* expel erecord/efield range control */
#define EXPEL_TYPE_N               0
#define EXPEL_TYPE_F               1
#define EXPEL_TYPE_C               2
#define RANGE_UNSET                -2

/* insert tag control */
#define TAG_INSERT_ELEMENT         0
#define TAG_INSERT_FILE            1
#define TAG_INSERT_XMLFILE         2

/* option identifying control */
#define OPT_DELIMITER              0
#define OPT_ATTRIBUTE              1
#define OPT_TAG                    2
#define OPT_COLUMN                 3
#define OPT_REGEX                  4

/* quote control */
#define QUOTE_CLOSE                0
#define QUOTE_OPEN                 1

/* summary control */
#define SUMMARY_NORMAL             1
#define SUMMARY_COMMENT            2
#define SUMMARY_FILE               3

/* end of line control */
#define NEWLINE_UNIX               "\n"
#define NEWLINE_DOS                "\r\n"

/* temp array control */
#define TEMP_ARRAY_SIZE            60

/* platform specific open() IO flags */
#ifndef XMLFY_WIN32
#define OPEN_R O_RDONLY
#define OPEN_W O_WRONLY|O_CREAT|O_TRUNC
#else
#define OPEN_R O_RDONLY|O_BINARY
#define OPEN_W O_WRONLY|O_CREAT|O_TRUNC|O_BINARY
#endif /* XMLFY_WIN32 */

extern const char progname[];

/* match helper structure */
typedef struct {
  int  min;             /* the minimum number of fields for a match */
  int  max;             /* the maximum number of fields for a match */
  int  last;            /* number of fields required after last wildcard */
  int  *xpath;          /* array of branch path traversal sequences */
} struct_match_helper;

/* element helper structure */
typedef struct {
  int  nxpaths;         /* number of xpath combinations */
  int  max_xpath_seq;   /* longest xpath sequence */
  int  complete;        /* switch to stop duplicate processing */
  struct_match_helper *match_helper;
} struct_element_helper;

/*
 * Schema structure layout:
 * One to many relationship between parent and child objects.
 *
 * <parent>
 *    |
 *    +=><element> or <field>....
 *          |
 *          +=><record>....
 *                |
 *                +=><element> or <field>....
 *                      |
 *                      +=><record>....
 *                            :
 *                            :
 */

/* Schema field structure */
typedef struct {
  unsigned char *name;  /* field name */
  int  type;            /* FIELD_CDATA or FIELD_SUBRECORD */
  int  match_control;   /* wildcard token */
  int  field_count;     /* helper for wildcard token processing */
} struct_schema_field;

/* Schema record structure */
typedef struct {
  unsigned char *name;  /* record name */
  int  record_count;    /* helper for wildcard token processing */
  int  nfields;         /* the number of fields of the record */
  struct_schema_field *field; /* the fields of the record */
} struct_schema_record;

/* Schema element structure */
typedef struct {
  unsigned char *name;  /* element name (pointer to first record name) */
  int  element_count;   /* helper for processing */
  int  nrecords;        /* the number of records of the element */
  struct_schema_record **record;  /* the records of the element */
  struct_element_helper *element_helper;  /* the xpath helpers of the element */
} struct_schema_element;

/* Schema parent structure */
typedef struct {
  unsigned char *name;  /* parent name (pointer to first element name */
  int  nelements;       /* the number of elements in the schema file */
  int  nrecords;        /* the number of records in the schema file */
  int  nlevels;         /* the number of levels in the schema file */
  struct_schema_record *record; /* the first record of the first element */
} struct_schema_parent;

extern struct_schema_parent  schema_parent;
extern struct_schema_element *schema_element;
extern struct_schema_record  *schema_record;

extern unsigned char *schema_filename;
extern unsigned char *schema_system_identifier;
extern unsigned char *schema_xsd_target_ns;
extern unsigned char *schema_parent_name;
extern unsigned char *schema_line_name;
extern unsigned char *schema_linenumber;

extern unsigned char *default_schema_parent_name;
extern unsigned char *default_schema_line_name;
extern unsigned char *default_schema_field_name;

/* level handling */
extern int nlevels;

/* switches */
extern int sw_xml_format;
extern int sw_encoding;
extern int sw_schema;
extern int sw_printonly_header;
extern int sw_printonly_footer;
extern int sw_printonly_rtagopen;
extern int sw_printonly_rtagclose;
extern int sw_printonly_records;
extern int sw_printonly_first_opt;
extern int sw_char_escape_amp;
extern int sw_char_escape_lt;
extern int sw_char_escape_gt;
extern int sw_char_escape_quot;
extern int sw_char_escape_apos;
extern int sw_char_escape_brvbar;
extern int sw_process_output_digest;
extern int sw_digest;
extern int sw_line_tags;
extern int sw_field_quoting;
extern int sw_field_quoting_span;
extern int sw_field_expel_empty;
extern int sw_field_empty_valid;
extern int sw_field_trim;
extern int sw_blank_lines;
extern int sw_key_value_pairs;
extern int sw_kvp_levels_all;
extern int sw_bom;
extern int sw_ascii_to_utf;
extern int sw_delimiter_scoping;
extern int sw_trimtagclose;
extern int sw_summary;
extern int sw_debug;
typedef struct {
  unsigned char *tagname; /* pointer to tagname string */
  int tagname_length;     /* length of tagname */
  int numbered_attr;      /* print number attributes */
  int numbered_tag;       /* print number tags */
  int delimiter_attr;     /* print delimiter attributes */
  int level_attr;         /* print level attributes */
  int level_tag;          /* print level tags */
  int timestamp_attr;     /* print timestamp attributes */
} struct_sw_print;
extern struct_sw_print **sw_print, sw_print_default;

/* key/value pair handling */
typedef struct {
  int enabled;          /* switch to invoke kvp for this level */
  int set;              /* switch to only invoke kvp once per level */
  unsigned char *name;  /* pointer to tagname string */
  int length;           /* length of tagname */
  int encoding;         /* encoding of tagname string */
  int level_d;          /* index to matching level of delimiter */
  int match_d;          /* index to matching delimiter */
} struct_kvp_tag;
extern struct_kvp_tag **kvp_tag;

/* IO handling */
extern unsigned char utf_buf[BUFSIZE];
extern int utf_flag_EOF;
extern int utf_read;
extern int utf_cntr;

/* input field structure */
typedef struct {
  int  level;           /* the XML level of this field */
  long sequence;        /* the sequence of this field in the current XML level */
  long start;           /* index to start-of-field in input_record.line array */
  long length;          /* index to end-of-field in input_record.line array */
  int  level_d;         /* index to matching level of delimiter */
  int  match_d;         /* index to matching delimiter */
  int  type;            /* FIELD_CDATA or FIELD_SUBRECORD */
  int  *schema_xpath;   /* array of indexes pointing to schema_element[]s */
  unsigned char *tagname; /* pointer to the XML tag name for this field */
} struct_input_field;

/* input record structure */
typedef struct {
  unsigned char *line;  /* the input record */
  long length;          /* the length of the input record */
  int  charwidth;       /* the minimum code point encoding size in bytes */
  int  encoding;        /* the encoding used in the input */
  unsigned char bom_name[ENCODING_NAME_SIZE]; /* the encoding name of the BOM if any */
  long nfields;         /* the total number of fields found in the input record */
  long nfields_cdata;   /* the number of FIELD_CDATA fields found in the input record */
  struct_input_field **field; /* the field allocations */
} struct_input_record;
extern struct_input_record input_record;

/* scope range structure */
typedef struct {
  long min;             /* minima of range */
  long max;             /* maxima of range */
  int  restart;         /* flag to restart scope counter */
} struct_scope;

/* delimiter structure */
typedef struct {
  long p1;               /* absolute first byte of comparison */
  long p2;               /* absolute last byte of delimiter */
  long p3;               /* absolute last byte of comparison */
  int match;             /* switch for match handling */
  int match_d;           /* index to matching delimiter */
  int ndelimiters;       /* number of delimiters */
  unsigned char **utf_d; /* delimiter strings in UTF format */
  int *utf_length;       /* corresponding length of each UTF delimiter string */
  unsigned char **arg_d; /* delimiter strings from args */
  int *arg_length;       /* corresponding length of each arg delimiter string */
  int *byte_mode;        /* switch to prevent conversion of arg delimiter to utf encoding */
  int *nscope;           /* corresponding number of scope options of each delimiter string */
  struct_scope **scope;  /* corresponding scope ranges of each delimiter string */
  long *scope_seq;       /* corresponding scope sequence of each delimiter string */
} struct_delimiter;
extern struct_delimiter **delimiter;
extern unsigned char delimiter_newline[2];
extern unsigned char delimiter_space[2];
extern unsigned char delimiter_tab[2];
extern unsigned char delimiter_cr[2];
extern int delimiter_byte_mode;

/* counter for number of expel structures allocated */
extern int nexpel;

/* expel records range structure */
typedef struct {
  int  type;            /* type of range */
  long min;             /* minima of range */
  long max;             /* maxima of range */
  unsigned char *s;     /* search string */
  int  s_length;        /* length of search string */
} struct_erecords;

/* expel fields range structure */
typedef struct {
  long min;             /* minima of range */
  long max;             /* maxima of range */
} struct_efields;

/* expel records/fields structure */
typedef struct {
  int             n_erecords; /* number of erecords structures */
  struct_erecords *erecords;  /* array of erecords */
  int             n_efields;  /* number of efields structures */
  struct_efields  *efields;   /* array of efields */
} struct_expel;
extern struct_expel *expel;

/* counter for number of column structures allocated */
extern int ncolumn;
extern int flag_input_has_composites; /* optimisation flag for column range searches */

/* column range/regex structure */
typedef struct {
  int type;               /* type of column - range or regex */
  unsigned char *name;    /* tagname of range or tagarray of regex arg */
  /* range members */
  long min;               /* minima of range */
  long max;               /* maxima of range */
  /* regex members */
  unsigned char *pattern; /* pointer to regex pattern option */
  int     pattern_length; /* length of regex pattern option */
  regex_t    pcomp;       /* compiled regex pattern */
  regmatch_t *pmatch;     /* regex pattern match array */
  int        eflags;      /* regex flags to regexec() */
  int        ntagname;    /* number of tagnames */
  unsigned char **tagname; /* tagname for each regex match */
  /* scope members */
  int nscope;             /* number of scopes for regex option */
  struct_scope *scope;    /* scope ranges of regex option */
  long scope_seq;         /* scope sequence of column field */
} struct_column;
extern struct_column *column;

/* counter for number of insert_attr structures allocated */
extern int ninsert_attr;

/* custom element attribute structure */
typedef struct {
  int  level;           /* level of this custom attribute */
  unsigned char *name;  /* name of custom attribute */
  int  name_length;     /* name length of custom attribute */
  unsigned char *value; /* value of custom attribute */
  int  value_length;    /* value length of custom attribute */
  int  nscope;          /* number of scopes for custom attribute */
  struct_scope *scope;  /* scope ranges of custom attribute */
  long scope_seq;       /* scope sequence of custom attribute */
} struct_insert_attr;
extern struct_insert_attr *insert_attr;

/* counter for number of insert_tag structures allocated */
extern int ninsert_tag;

/* custom element tag structure */
typedef struct {
  int  type;                  /* type of custom tag */
  int  level;                 /* level of this custom tag */
  unsigned char *name;        /* name of custom custom tag */
  int  name_length;           /* name length of custom tag */
  unsigned char *value;       /* value of custom custom tag */
  long value_length;          /* value length of custom tag */
  unsigned char *filename;    /* value of filename */
  int  indent;                /* indentation for this custom tag */
  int  reinsert;              /* insert mode of custom tag */
  int  nscope;                /* number of scopes for custom tag */
  struct_scope *scope;        /* scope ranges of custom tag */
  long scope_seq;             /* scope sequence of custom tag */
} struct_insert_tag;
extern struct_insert_tag *insert_tag;

/* counter for number of scope structures allocated */
extern int nscope;
extern struct_scope *scope;
/* marker for first level of restarting delimiter scopes */
extern int restarting_delimiter_scope_level;

/* output line handling */
typedef struct {
  unsigned char *line;  /* the output line */
  long length;          /* the length of the output line */
  int  charwidth;       /* the minimum code point encoding size in bytes */
  int  encoding;        /* the encoding used in the output */
  unsigned char encoding_name[ENCODING_NAME_SIZE]; /* the encoding name used in the prologue */
  int *field_cntr;      /* the field counter */
} struct_output_record;
extern struct_output_record output_record;

/* output newline handling */
extern unsigned char newline[3];

/* utf format handling function pointers */
typedef void (*fptr_void)();
extern fptr_void strcpy_output_record;
extern fptr_void strncpy_output_record;
extern fptr_void strcat_output_record;
extern fptr_void strncat_output_record;

/* memory blocks handling */
extern long memseg_nfields;
extern long memseg_input_record;
extern long memseg_output_record;
/* memory allocation function call counters */
extern int  memadd_nfields_cntr;
extern int  memadd_input_record_cntr;
extern int  memadd_output_record_cntr;

/* timestamp handling */
extern time_t start_ts, ts;

/* schema match start level handling */
extern int schema_match_start_level; /* The level to start record matching */ 
extern unsigned char *schema_match_start_name;

/* schema file option handling */
extern unsigned char **schema_argv;
extern int schema_argc;

/* quotechar handling */
extern int quotechar_type;        /* quotechar type 0=default 1=array 2=paired */
extern unsigned char *quotechars; /* pointer to quotechar array arg */
extern int quotechars_length;     /* length of quotechar array arg */
extern int i_quotechar;           /* iterator for processing quotechar array */

/* XML/SOAP version implementation handling */
extern int version_xml;
extern int version_soap;

/* summary handling */
extern int summary_fd;

/* MD5 signature handling */
extern MD5_CTX md5_in_context;
extern MD5_CTX md5_out_context;

/* SHA512 signature handling */
extern SHA512Context sha512_in_context;
extern SHA512Context sha512_out_context;

/* HTML output handling */
extern unsigned char *html_parent_tagname;
extern unsigned char *html_child_tagname;

/* function templates */
extern void validate_args(int argc, unsigned char **argv, int invocation);
extern void schema();
extern void xmlout(int argc, unsigned char **argv);
extern void utf_print(int fd, unsigned char *buf, long buf_bytes);
extern unsigned char utf_getchar(int fd);
extern void utf_getchar_init();
extern void noxml();
extern void unxml();
extern void memadd_nfields();
extern void memadd_input_record();
extern void memadd_output_record(long msize);
extern void out_of_mem();
/* xmlfy simplified libc like routines to handle unsigned char */
extern void xstrcpy(unsigned char *s_dst, const unsigned char *s_src);
extern void xstrncpy(unsigned char *s_dst, const unsigned char *s_src, long n);
extern int  xstrcmp(const unsigned char *s1, const unsigned char *s2);
extern int  xstrncmp(const unsigned char *s1, const unsigned char *s2, long n);
extern int  xstrlen(const unsigned char *s1);

/* error status handling */
extern int err_status;

#endif /* XMLFY_H */
