/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlout.h"

/* function templates */
int match_expel(long start, long end);
int expel_empty_fields();


int match_expel(start, end)
long start;
long end;
{
  /* match any records and fields against the expel criteria and expel them from processing */
  int i, j, k;
  long m, n, f_cntr, temp_min, temp_max;
  int flag_expel;

  flag_expel = EXPEL_NONE;

  /* expel empty fields */
  if (sw_field_expel_empty)
    flag_expel = expel_empty_fields();

  for (i = 0 ; i < nexpel && flag_expel == EXPEL_NONE ; i++)
  {
    flag_expel = EXPEL_RECORD;
    for (j = 0 ; j < expel[i].n_erecords && flag_expel == EXPEL_RECORD ; j++)
    {
      /* check against different expel criteria */
      switch(expel[i].erecords[j].type)
      {
        case EXPEL_TYPE_N:
          if (expel[i].erecords[j].s_length < 1)
          {
            if (record_cntr_read < expel[i].erecords[j].min || (record_cntr_read > expel[i].erecords[j].max && expel[i].erecords[j].max >= 0))
              flag_expel = EXPEL_NONE;
          }
          else
          {
            /* do a string search */
            flag_expel = EXPEL_NONE;
            for (k = start ; k < end ; k += input_record.charwidth)
            {
              if (utf_xstrncmp(&input_record.line[k], end, input_record.encoding, expel[i].erecords[j].s, expel[i].erecords[j].s_length) == 0)
              {
                flag_expel = EXPEL_RECORD;
                break;
              }
            }
          }
          break;

        case EXPEL_TYPE_F:
          f_cntr = -1;
          for (m = 0 ; m < input_record.nfields ; m++)
            if (input_record.field[m]->type == FIELD_CDATA)
              f_cntr++;
          if (expel[i].erecords[j].s_length < 1)
          {
            if (f_cntr < expel[i].erecords[j].min || (f_cntr > expel[i].erecords[j].max && expel[i].erecords[j].max >= 0))
              flag_expel = EXPEL_NONE;
          }
          else
          {
            /* do a string search */
            flag_expel = EXPEL_NONE;
            f_cntr = -1;
            for (m = 0 ; m < input_record.nfields && flag_expel == EXPEL_NONE ; m++)
            {
              if (input_record.field[m]->type == FIELD_CDATA)
                f_cntr++;
              if (f_cntr >= expel[i].erecords[j].min && (f_cntr <= expel[i].erecords[j].max || expel[i].erecords[j].max < 0))
              {
                for (k = input_record.field[m]->start ; k < input_record.field[m]->start + input_record.field[m]->length ; k += input_record.charwidth)
                {
                  if (utf_xstrncmp(&input_record.line[k], input_record.field[m]->length, input_record.encoding, expel[i].erecords[j].s, expel[i].erecords[j].s_length) == 0)
                  {
                    flag_expel = EXPEL_RECORD;
                    break;
                  }
                }
              }
            }
          }
          break;

        case EXPEL_TYPE_C:
          /* Adjust min and max values for larger sized code points.
           * E.g. Multi byte UTF-8 sequences.
           * E.g. Surrogate pairing in UTF-16.  */
          temp_min = start + (expel[i].erecords[j].min * input_record.charwidth);
          if (expel[i].erecords[j].max > -1)
            temp_max = start + ((expel[i].erecords[j].max + 1) * input_record.charwidth);
          else
            temp_max = end;

          for (m = start ; m < end ; m += input_record.charwidth)
          {
            if (utf_iscomposite(m))
            {
              if (m < temp_min)
              {
                temp_min += input_record.charwidth;
                temp_max += input_record.charwidth;
              }
              else if (m < temp_max)
                temp_max += input_record.charwidth;
              else
                break; /* for/m */
            }
          }

          /* boundary checks */
          if (temp_min < start)
            temp_min = start;
          if (temp_min > end)
            temp_min = end;
          if (temp_min < 0)
            temp_min = 0;
          if (temp_max > end)
            temp_max = end;
          if (temp_min > temp_max)
            temp_min = temp_max;

          if (expel[i].erecords[j].s_length < 1) 
          {
            if (end - start < temp_min + 1 || (end - start > temp_max + 1 && expel[i].erecords[j].max > -1))
              flag_expel = EXPEL_NONE;
          }
          else
          {
            /* do a string search between code-points rather than bytes */
            flag_expel = EXPEL_NONE;
            for (k = temp_min ; k < temp_max ; k += input_record.charwidth)
            {
              if (utf_xstrncmp(&input_record.line[k], m, input_record.encoding, expel[i].erecords[j].s, expel[i].erecords[j].s_length) == 0)
              {
                flag_expel = EXPEL_RECORD;
                break;
              }
            }
          }
      } /* end switch */
    } /* end for/j */

    if (flag_expel == EXPEL_RECORD)
    {
      if (expel[i].n_efields)
      {
        flag_expel = EXPEL_FIELDS;
        f_cntr = 0;
        for (m = 0 ; m < input_record.nfields ; m++)
        {
          if (input_record.field[m]->type == FIELD_CDATA)
          {
            for (k = 0 ; k < expel[i].n_efields ; k++)
            {
              temp_min = expel[i].efields[k].min;
              if (expel[i].efields[k].min < 0)
              {
                /* figure out the last field */
                temp_min = f_cntr;
                for (n = m + 1 ; n < input_record.nfields ; n++)
                  if (input_record.field[n]->type == FIELD_CDATA)
                    temp_min++;
              }
              if (f_cntr >= temp_min && (f_cntr <= expel[i].efields[k].max || expel[i].efields[k].max < 0))
                input_record.field[m]->type = FIELD_EXPELLED;
            } /* end for/k */
            f_cntr++;
          }
        } /* end for/m */
      }
    }
  } /* end for/i */

  if (flag_expel == EXPEL_RECORD)
    record_cntr_expelled++;

  return(flag_expel);
}


int expel_empty_fields()
{
  /* expel empty fields from the input record */
  long i, p1, p2;
  int current_level;
  int flag_expel_record;

  /* initialise */
  p1 = 0;
  p2 = 0;
  current_level = 2;
  /* set flag to expel records made up entirely of empty fields */
  flag_expel_record = EXPEL_RECORD; 

  while (p1 < input_record.nfields)
  {
    /* locate non-empty field */
    for (p2 = p1 ; p2 < input_record.nfields ; p2++)
    {
      if (input_record.field[p2]->type == FIELD_CDATA && input_record.field[p2]->length)
      {
        current_level = input_record.field[p2]->level;
        /* record has non-empty fields */
        flag_expel_record = EXPEL_NONE;
        break;
      }
    }
    if (p2 >= input_record.nfields)
      p2 = input_record.nfields - 1;

    /* Key/value pairs mode handles blank fields in its own way */
    if (!sw_kvp_levels_all)
    {
      /* work backwards expelling any empty input fields and the subrecords that created them */
      for (i = p2 ; i >= p1 ; i--)
      {
        if (input_record.field[i]->level <= current_level)
        {
          if (!input_record.field[i]->length)
          {
            if (!sw_key_value_pairs || !kvp_tag[input_record.field[i]->level]->enabled)
              input_record.field[i]->type = FIELD_EXPELLED;
          }
          current_level = input_record.field[i]->level;
        }
        else if (!sw_key_value_pairs || !kvp_tag[input_record.field[i]->level]->enabled)
        {
          input_record.field[i]->type = FIELD_EXPELLED;
          /* knock down empty fields with superfluous higher level */
          input_record.field[i]->level = current_level;
        }
      } 
    }
    else
      break; /* while */
    p1 = ++p2;
  } /* end while */

  return(flag_expel_record);
}

