/*
 * BSD License for xmlfy
 * Copyright (c) 2008-2013, Arthur Gouros
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * - Neither the name of Arthur Gouros nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**************************************************************************
 * xmlfy - Convert to XML on the fly.                                     *
 *         Original Author: Arthur Gouros.                                *
 *                                                                        *
 *         Please consult the documentation for further information.      *
 **************************************************************************
 */

#include "xmlout.h"

int flag_spanning_quote;

/* function templates */
int  process_field(int level, long pos, long start, long end);
void process_column_fields(long start, long end);
void process_column_range_fields(int i, long start, long end);
void process_column_regex_fields(int i, long start, long end);
void process_record(long start, long end);
int  delimiter_separator(int level, long pos, long end, int buffer_size, int depth);
int  check_delimiter_scope(int level, int i);
void reset_delimiter_scope_seq(int level);
void reset_restarting_delimiter_scope_seq(int level, int i, int j);
int  quote_separator(long i, int match_fwd);
void trim_field(long i);
void recycle_field(long i);


void process_input()
{
  /* process stdin one record at a time and convert it to XML format */
  int buffer_size;

  /* initialise */
  delimiter[1]->p1 = 0;
  delimiter[1]->p2 = 0;
  delimiter[1]->p3 = 0;
  delimiter[1]->match_d = DELIMITER_INDEX_UNMATCHED;
  if (delimiter[1]->ndelimiters > 0)
    buffer_size = delimiter[1]->utf_length[0];
  else
    buffer_size = 0;

  flag_spanning_quote = FALSE;

  /* process input */
  while (!utf_flag_EOF)
  {
    input_record.length = 0;
    input_record.nfields = 0;

    /* put unmatched delimiter data back into readback buffer */
    if (delimiter[1]->match)
    {
      while(delimiter[1]->p2 < delimiter[1]->p3)
        input_record.line[input_record.length++] = input_record.line[++delimiter[1]->p2];
      delimiter[1]->p1 = 0;
      delimiter[1]->p2 = 0;
      delimiter[1]->p3 = 0;
    }

    /* read record into array then break into fields */
    do 
    {
      input_record.line[input_record.length++] = utf_getchar(STDIN_FD);

      /* mem checks */
      if (input_record.length > memseg_input_record)
        memadd_input_record();

    } while (!utf_flag_EOF && !delimiter_separator(1, input_record.length - 1, input_record.length - 1, buffer_size, 0));

    /* locate and breakdown fields of input record */
    if (!utf_flag_EOF)
    {
        if (ncolumn)
          process_column_fields(0, delimiter[1]->p1);
        else
        {
          if (delimiter[1]->p1 > 0 || sw_blank_lines)
            (void) process_field(2, 0, 0, delimiter[1]->p1);
        }
        process_record(0, delimiter[1]->p1);
    }
  }

  /* do-while logic adds EOF so ignore it */
  if (input_record.length > 0)
    input_record.length--; 

  /* Clarification when using the default level 1 delimiter NEWLINE:
   * Most Unix text files finish their last line with a newline character.
   * By default, don't process the very last character if it's a NEWLINE
   * in order to not trigger a new empty record.
   * If you really must process this last NEWLINE character as a record
   * delimiter then use the explicit option -F1 "\n" 
   */
  if (delimiter[1]->ndelimiters == 0 && delimiter[1]->p2 + 1 == input_record.length && input_record.line[delimiter[1]->p2] == NEWLINE)
    input_record.length--; 

  if (input_record.length > 0 || flag_spanning_quote)
  {
    /* turn off in case we encountered EOF and unclosed quotes */
    flag_spanning_quote = FALSE;
   
    /* process the readback buffer.
       This must be done at level 1 in case delimiters are present */
    if (input_record.length > 0 || sw_blank_lines)
      (void) process_field(1, input_record.length - 1 - buffer_size, 0, input_record.length);
  }
}


int process_field(level, pos, start, end)
int level;
long pos;
long start;
long end;
{
  /* look for sub fields in field segment recursively.
   * I concur this function aint pretty */
  int depth, match_cntr;
  long i, start_next, delimiter_next, current_nfields;
  int flag_process_more, flag_quoted_field;

  /* initialise */
  match_cntr        = 0;
  flag_quoted_field = FALSE;
  start_next        = start;
  current_nfields   = 0;
  delimiter_next    = pos;
  if (delimiter_next < 0) 
    delimiter_next = 0;

  delimiter[level]->match_d = DELIMITER_INDEX_UNMATCHED;
  delimiter[level]->p1 = 0;
  delimiter[level]->p2 = 0;
  depth = 0;

  /* if field delimiter scoping is used then reset counters */
  if (sw_delimiter_scoping)
    reset_delimiter_scope_seq(level);

  for (i = delimiter_next ; i < end ; i += input_record.charwidth)
  {
    /* process quoted field */
    if (level > 1 && quote_separator(i, TRUE))
    {
      flag_quoted_field = TRUE;
      i += input_record.charwidth;
      input_record.field[input_record.nfields]->level   = level;
      input_record.field[input_record.nfields]->start   = i;
      input_record.field[input_record.nfields]->length  = 0;
      input_record.field[input_record.nfields]->level_d = level;
      input_record.field[input_record.nfields]->match_d = delimiter[level]->match_d;
      input_record.field[input_record.nfields]->type    = FIELD_CDATA;
      input_record.field[input_record.nfields]->tagname = sw_print[level]->tagname;
      while (i <= end && !quote_separator(i, TRUE))
      {
        input_record.field[input_record.nfields]->length += input_record.charwidth;
        i += input_record.charwidth;
      }
      /* not processing this allows for delimiter alignment 
        if (quote_separator(i, TRUE) && i < end)
         i += input_record.charwidth;
      */
      current_nfields = input_record.nfields;
      input_record.nfields++;

      if (level + 1 < nlevels && (delimiter[level + 1]->ndelimiters || level <= 2))
      {
        if (process_field(level + 1, input_record.field[current_nfields]->start, input_record.field[current_nfields]->start, input_record.field[current_nfields]->start + input_record.field[current_nfields]->length))
        {
          input_record.field[current_nfields]->type = FIELD_SUBRECORD;
          if (sw_schema != SCHEMA_DEFAULT)
            recycle_field(current_nfields);
        }
      }
      if (sw_field_trim && input_record.field[current_nfields]->type == FIELD_CDATA)
        trim_field(current_nfields);
    }
    else
    {
      while (depth < delimiter[level]->ndelimiters && delimiter[level]->utf_length[depth] > end - i)
        depth++;
      if (delimiter_separator(level, i, end, end - i, depth))
      {
        match_cntr++;
        flag_process_more = TRUE;
        if (level == 1 && !sw_blank_lines && delimiter[level]->p1 - start_next == 0)
          flag_process_more = FALSE;
        if (level > 1 && !flag_quoted_field)
        {
          input_record.field[input_record.nfields]->level   = level;
          input_record.field[input_record.nfields]->start   = start_next;
          input_record.field[input_record.nfields]->length  = delimiter[level]->p1 - start_next;
          input_record.field[input_record.nfields]->level_d = level;
          input_record.field[input_record.nfields]->match_d = delimiter[level]->match_d;
          input_record.field[input_record.nfields]->type    = FIELD_CDATA;
          input_record.field[input_record.nfields]->tagname = sw_print[level]->tagname;
          current_nfields = input_record.nfields;
          if (input_record.field[current_nfields]->length > 0 || sw_field_empty_valid)
            input_record.nfields++;
        }

        if ((level + 1 < nlevels && (delimiter[level + 1]->ndelimiters || level <= 2)) && !flag_quoted_field && flag_process_more)
        {
          if (process_field(level + 1, start_next, start_next, delimiter[level]->p1))
          {
            if (level > 1)
            {
              input_record.field[current_nfields]->type = FIELD_SUBRECORD;
              if (sw_schema != SCHEMA_DEFAULT)
                recycle_field(current_nfields);
            }
          }
        }
        if (sw_field_trim && input_record.field[current_nfields]->type == FIELD_CDATA && !flag_quoted_field)
          trim_field(current_nfields);
        if (level == 1) 
        {
          if (ncolumn)
            process_column_fields(start_next, delimiter[1]->p1);
          process_record(start_next, delimiter[1]->p1);
        }

        /* update to end of matching delimiter */
        start_next = delimiter[level]->p2 + 1;

        /* process to next delimiter */
        flag_quoted_field = FALSE;
      }
    }
    /* mem checks */
    if (input_record.nfields > memseg_nfields)
      memadd_nfields();
  } 

  /* process last undelimited data */
  delimiter[level]->match_d = DELIMITER_INDEX_UNMATCHED;
  if (end - start_next >= 0)
  {
    flag_process_more = TRUE;
    if (level == 1 && !sw_blank_lines && end - start_next == 0)
      flag_process_more = FALSE;
    if (level > 2 && start_next == start)
      flag_process_more = FALSE;
    if (level > 1 && !flag_quoted_field && flag_process_more)
    {
      match_cntr++;
      current_nfields = input_record.nfields;
      input_record.field[input_record.nfields]->level   = level;
      input_record.field[input_record.nfields]->start   = start_next;
      input_record.field[input_record.nfields]->length  = end - start_next;
      input_record.field[input_record.nfields]->level_d = level;
      input_record.field[input_record.nfields]->match_d = delimiter[level]->match_d;
      input_record.field[input_record.nfields]->type    = FIELD_CDATA;
      input_record.field[input_record.nfields]->tagname = sw_print[level]->tagname;
      if (input_record.field[current_nfields]->length > 0 || sw_field_empty_valid)
        input_record.nfields++;
    }
    if ((level + 1 < nlevels && (delimiter[level + 1]->ndelimiters || level <= 2)) && !flag_quoted_field && flag_process_more)
    {
      if (process_field(level + 1, start_next, start_next, end))
      {
        if (level > 1)
        {
          input_record.field[current_nfields]->type = FIELD_SUBRECORD;
          if (sw_schema != SCHEMA_DEFAULT)
            recycle_field(current_nfields);
        }
      }
    }
    if (sw_field_trim && input_record.field[current_nfields]->type == FIELD_CDATA && !flag_quoted_field)
      trim_field(current_nfields);
    if (level == 1) 
    {
      if (ncolumn && flag_process_more)
        process_column_fields(start_next, end);
      process_record(start_next, end);
    }
  }

  return(match_cntr);
}


void recycle_field(i)
long i;
{
  /* With schema processing all input fields are flattened to be just (level 2) fields.
   * Fields of type FIELD_SUBRECORD serve no purpose in this mode so reposition
   * them at the end of array for recycling and contiguous FIELD_CDATA processing. */
  struct_input_field *tmpfield;

  /* save field memory allocation */
  tmpfield = input_record.field[i];
  /* shift fields down */
  while (i < input_record.nfields - 1)
  {
    input_record.field[i] = input_record.field[i + 1];
    i++;
  }
  /* recycle saved field memory allocation */
  input_record.field[i] = tmpfield;

  /* re-use repositioned field memory allocation */
  if (input_record.nfields)
    input_record.nfields--;
}


void process_column_fields(start, end)
long start;
long end;
{
  /* process column range and column regex in order supplied at command line */
  int i, j;
  int flag_process_column;

  /* optimisation flag to prevent multiple range searching when no composite chars are present */
  flag_input_has_composites = TRUE;

  input_record.nfields = 0;
  for (i = 0 ; i < ncolumn ; i++)
  {
    flag_process_column = TRUE;

    /* check for scope restrictions */
    if (column[i].nscope)
    {
      flag_process_column = FALSE;
      for (j = 0 ; j < column[i].nscope ; j++)
      {
        if (column[i].scope_seq >= column[i].scope[j].min 
           && (column[i].scope[j].max < 0 || column[i].scope_seq <= column[i].scope[j].max))
        {
          flag_process_column = TRUE;
          break; /* break for/j */
        }
      } /* end for/j */

      /* increment scope sequence */
      column[i].scope_seq++;

      /* check for retarting scope and restart */
      if (flag_process_column && column[i].scope[j].restart && column[i].scope[j].max >= 0 && column[i].scope[j].max == column[i].scope_seq - 1)
        column[i].scope_seq = 0;
    }

    if (flag_process_column)
    {
      if (column[i].type == OPT_COLUMN)
        process_column_range_fields(i, start, end);
      else if (column[i].type == OPT_REGEX)
        process_column_regex_fields(i, start, end);
    }
  }
}


void process_column_range_fields(i, start, end)
int  i;
long start;
long end;
{
  /* Create input fields from column markers applied to level 1
   * input rcords.
   * NOTE: last column token ($) is represented as -1 
   */
  long j, p1, p2;
  
  /* mem checks */
  if (input_record.nfields > memseg_nfields)
    memadd_nfields();

  /* recalculate column positions in relative terms to start/end */
  if (column[i].min > -1)
    p1 = start + (column[i].min * input_record.charwidth);
  else
    p1 = end - input_record.charwidth;

  if (column[i].max > -1)
    p2 = start + ((column[i].max + 1) * input_record.charwidth);
  else
    p2 = end;

  /* Adjust p1 and p2 for larger sized code points.
   * E.g. Multi byte UTF-8 sequences.
   * E.g. Surrogate pairing in UTF-16.  */
  if (flag_input_has_composites)
  {
    flag_input_has_composites = FALSE;
    for (j = start ; j < end ; j += input_record.charwidth)
    {
      if (utf_iscomposite(j))
      {
        flag_input_has_composites = TRUE;
        if (j < p1)
        {
          p1 += input_record.charwidth;
          p2 += input_record.charwidth;
        }
        else if (j < p2 + input_record.charwidth)
          p2 += input_record.charwidth;
        else
          break; /* for/j */
      }
    }
  }

  /* Boundary check */
  if (p1 < start)
    p1 = start;
  if (p1 > end)
    p1 = end;
  if (p1 < 0)
    p1 = 0;
  if (p2 > end)
    p2 = end;
  if (p1 > p2)
    p1 = p2;

  /* set new input field attributes */
  input_record.field[input_record.nfields]->level   = 2;
  input_record.field[input_record.nfields]->start   = p1;
  input_record.field[input_record.nfields]->length  = p2 - p1;
  input_record.field[input_record.nfields]->level_d = input_record.nfields; /* actually points to a column array */
  input_record.field[input_record.nfields]->match_d = DELIMITER_INDEX_COLUMN;
  input_record.field[input_record.nfields]->type    = FIELD_CDATA;
  if (column[i].name == NULL)
    input_record.field[input_record.nfields]->tagname = sw_print[2]->tagname;
  else
    input_record.field[input_record.nfields]->tagname = column[i].name;

  if (sw_field_trim)
    trim_field(input_record.nfields);

  input_record.nfields++;
}


void process_column_regex_fields(i, start, end)
int  i;
long start;
long end;
{
  /* Create input fields from regex pattern syntax applied to level 1
   * input rcords. */
  int j, k, e, nmatch;
  
  k = 0;
  /* match input_record.line against RE */
  nmatch = column[i].pcomp.re_nsub + 1;

  /* The TRE doco is not clear whether pmatch[0] still has a REG_STARTEND meaning */
  /* column[i].pmatch[0].rm_so = start;
   * column[i].pmatch[0].rm_eo = end; 
   * Therefore assume it does not and work around it. */
  if ((e = regnexec(&column[i].pcomp, (char *) &input_record.line[start], end - start, nmatch, column[i].pmatch, column[i].eflags)) == 0)
  {
    for (j = 0 ; j < nmatch ; j++)
    {
      /* skip the pmatch[0] if parenthesis were used */
      if (j == 0 && nmatch > 1)
        continue;

      /* mem checks */
      if (input_record.nfields > memseg_nfields)
        memadd_nfields();

      /* set new input field attributes */
      input_record.field[input_record.nfields]->level = 2;
      if (nmatch > 1)
      {
        if (column[i].pmatch[j].rm_so < 0 || column[i].pmatch[j].rm_eo < 0)
        {
          /* No match defined when rm_so or rm_eo = -1 */
          input_record.field[input_record.nfields]->start  = 0;
          input_record.field[input_record.nfields]->length = 0;
        }
        else
        {
          input_record.field[input_record.nfields]->start  = start + column[i].pmatch[j].rm_so;
          input_record.field[input_record.nfields]->length = start + column[i].pmatch[j].rm_eo - column[i].pmatch[j].rm_so;
        }
      }
      else /* No parenthesis were specified */
      {
        input_record.field[input_record.nfields]->start  = start;
        input_record.field[input_record.nfields]->length = end - start;
      }
      input_record.field[input_record.nfields]->level_d = i; /* actually points to a regex pattern array */
      input_record.field[input_record.nfields]->match_d = DELIMITER_INDEX_REGEX;
      input_record.field[input_record.nfields]->type    = FIELD_CDATA;
      if (!column[i].ntagname)
        input_record.field[input_record.nfields]->tagname = sw_print[2]->tagname;
      else
      {
        input_record.field[input_record.nfields]->tagname = column[i].tagname[k];
        if (k + 1 < column[i].ntagname)
          k++;
      }
    
      if (sw_field_trim)
        trim_field(input_record.nfields);

      input_record.nfields++;
    }
  }
  else if (sw_debug)
  {
    j = regerror(e, &column[i].pcomp, temp_array, TEMP_ARRAY_SIZE);
    fprintf(stderr, "DEBUG: regexec(): input_record=%ld %s%s", record_cntr_read + 1, temp_array, newline);
  }
}


void process_record(start, end)
long start;
long end;
{
  /* process one record and convert it to XML format */

  if ((nexpel == 0 && !sw_field_expel_empty) || match_expel(start, end) != EXPEL_RECORD)
  {
    if (sw_schema == SCHEMA_DEFAULT)
      process_output(record_cntr_printed + 1);
    else
      match_schema();
  }

  record_cntr_read++;

  /* reset record for next pass */
  delimiter[1]->match_d = DELIMITER_INDEX_UNMATCHED;
  input_record.nfields = 0;
}


int quote_separator(i, match_fwd)
long i;
int  match_fwd;
{
  /* check for quoting control */
  int ret = FALSE;

  if (sw_field_quoting)
  { 
    if (quotechars_length)
    {
      if (quotechar_type == 1) /* check against array of quote characters */
      {
        /* if quote array supplied iterate through the array */
        for (i_quotechar = 0 ; i_quotechar < quotechars_length ; i_quotechar++)
        {
          if (match_fwd)
          {
            if (utf_char_equals(i, quotechars[i_quotechar]))
            {
              ret = TRUE;
              break;  /* for/i_quotechar */
            }
          }
          else 
          {
            if (utf_char_equals_level1(i, quotechars[i_quotechar]))
            {
              ret = TRUE;
              break;  /* for/i_quotechar */
            }
          }
        } /* end for/i_quotechar */
      }
      else if (quotechar_type == 2) /* check against pairs inside array of quote characters */
      {
        if (i_quotechar) /* quote is open, check for matching closed quote pair */
        {
          if (match_fwd)
          {
            if (utf_char_equals(i, quotechars[i_quotechar]))
            {
              i_quotechar = 0; /* quote is now closed, so reset */
              ret = TRUE;
            }
          }
          else 
          {
            if (utf_char_equals_level1(i, quotechars[i_quotechar]))
            {
              i_quotechar = 0; /* quote is now closed, so reset */
              ret = TRUE;
            }
          }
        }
        else /* quote is not open, iterate through open quote pairs for a match */
        {
          for (i_quotechar = 0 ; i_quotechar < quotechars_length ; i_quotechar += 2)
          {
            if (match_fwd)
            {
              if (utf_char_equals(i, quotechars[i_quotechar]))
              {
                i_quotechar += 1; /* re-point to matching closed pair */
                ret = TRUE;
                break;  /* for/i_quotechar */
              }
            }
            else 
            {
              if (utf_char_equals_level1(i, quotechars[i_quotechar]))
              {
                i_quotechar += 1; /* re-point to matching closed pair */
                ret = TRUE;
                break;  /* for/i_quotechar */
              }
            }
          } /* end for/i_quotechar */
          if (!ret)
            i_quotechar = 0; /* quote is still not open, so reset */
        }
      }
    }
    else if (utf_char_equals(i, QUOTECHAR))
      ret = TRUE;
  }

  return(ret);
}


void trim_field(i)
long i;
{
  /* trim the field of whitespace from either side */

  /* remove leading trim */
  while (input_record.field[i]->length > 0 && utf_isspace(input_record.field[i]->start))
  {
    input_record.field[i]->length -= input_record.charwidth;
    input_record.field[i]->start  += input_record.charwidth;
  }
  /* remove trailing trim */
  while (input_record.field[i]->length > 0 && utf_isspace(input_record.field[i]->start + input_record.field[i]->length - input_record.charwidth))
    input_record.field[i]->length -= input_record.charwidth;
}


int delimiter_separator(level, pos, end, buffer_size, depth)
int level;
long pos;
long end;
int buffer_size;
int depth;
{
  /* check to see if matched on record separators */
  int i, j;

  delimiter[level]->match = FALSE;

  if (sw_field_quoting_span && level == 1 && quote_separator(pos, FALSE))
  {
    if (flag_spanning_quote)
      flag_spanning_quote = FALSE;
    else
      flag_spanning_quote = TRUE;
  }

  if (flag_spanning_quote)
  {
    /* do nothing because already set delimiter[level]->match = FALSE */
    ;
  }
  else if (delimiter[level]->ndelimiters)
  {
    /* length[level] is sorted in decreasing size */
    if (delimiter[level]->utf_length[depth] <= buffer_size && end - buffer_size >= 0) 
    {
      /* p1 = absolute first byte of comparison */
      delimiter[level]->p1 = end - buffer_size;
      /* iterate through delimiters */
      for (i = depth ; i < delimiter[level]->ndelimiters && !delimiter[level]->match ; i++)
      {
        for (j = 0 ; j < delimiter[level]->utf_length[i] && 
            delimiter[level]->utf_d[i][j] == input_record.line[delimiter[level]->p1 + j] ; j++)
          ;
        if (j == delimiter[level]->utf_length[i] && check_delimiter_scope(level, i)) 
        {
          delimiter[level]->match = TRUE;
          delimiter[level]->match_d = i;
          /* save locations for processing in next pass */
          /* p2 = absolute last byte of delimiter */
          delimiter[level]->p2 = delimiter[level]->p1 + j - 1;
          /* p3 = absolute last byte of comparison */
          delimiter[level]->p3 = pos;
        }
      }
    }
  }
  else /* using default delimiter */
  {
    if (level == 1)
    {
      /* default is NEWLINE */
      if (utf_char_equals_level1(pos, NEWLINE)) 
      {
        delimiter[1]->match   = TRUE;
        delimiter[1]->p1      = pos - input_record.charwidth + 1;
        delimiter[1]->p2      = pos;
        delimiter[1]->p3      = pos;
        delimiter[1]->match_d = DELIMITER_INDEX_NEWLINE;
      }
    }
    else if (level == 2)
    {
      /* default is WHITESPACE */
      if (utf_isspace(pos))
      {
        delimiter[2]->match   = TRUE;
        delimiter[2]->p1      = pos;
        delimiter[2]->p2      = pos + input_record.charwidth - 1;
        delimiter[2]->p3      = pos + input_record.charwidth - 1;
        delimiter[2]->match_d = DELIMITER_INDEX_WHITESPACE;
      }
    }
  }

  return(delimiter[level]->match);
}


int check_delimiter_scope(level, i)
int level;
int i;
{
  /* check if delimiter has scope restrictions */
  int j = 0, ret = TRUE;

  if (delimiter[level]->nscope[i])
  {
    ret = FALSE;
    for (j = 0 ; j < delimiter[level]->nscope[i] ; j++)
    {
      if (delimiter[level]->scope_seq[i] >= delimiter[level]->scope[i][j].min 
          && (delimiter[level]->scope[i][j].max < 0 || delimiter[level]->scope_seq[i] <= delimiter[level]->scope[i][j].max))
      {
        ret = TRUE;
        break; /* break for/j */
      }
    }
    /* increment scope sequence */
    delimiter[level]->scope_seq[i]++;
  }

  /* Reset any restarting scopes. Optimised for performance and a bit of ugliness */
  if (restarting_delimiter_scope_level)
  {
    if (ret && level >= restarting_delimiter_scope_level)
      reset_restarting_delimiter_scope_seq(level, i, j);
  }

  return(ret);
}


void reset_delimiter_scope_seq(level)
int level;
{
  /* reset the delimiter scope sequence counter for current level and higher */
  int i;

  while (level < nlevels)
  {
    for (i = 0 ; i < delimiter[level]->ndelimiters ; i++)
      delimiter[level]->scope_seq[i] = 0;
    level++;
  }
}


void reset_restarting_delimiter_scope_seq(level, i, j)
int level;
int i;
int j;
{
  /* reset the delimiter scope sequence counter for current level and higher
   * but only for restarting scopes */
  int m, n;
  int flag_restart_scope = FALSE;

  if (!delimiter[level]->nscope[i])
  {
    /* Delimiter has no scope options so restart counters */
    flag_restart_scope = TRUE;
  }
  else /* one or many scope options */
  {
    if (j < delimiter[level]->nscope[i] && delimiter[level]->scope[i][j].restart)
    {
      /* Current delimiter's scope requests a counter restart */
      flag_restart_scope = TRUE;
    }
    else
    {
      /* Assume scope with no restart option */
      flag_restart_scope = TRUE;
      /* Now scan through delimiter's scope options looking for a
       * restart option elsewhere */
      for (n = 0 ; n < delimiter[level]->nscope[i] ; n++)
        if (delimiter[level]->scope[i][n].restart)
          flag_restart_scope = FALSE;
    }
  }

  if (flag_restart_scope)
  {
    while (level < nlevels)
    {
      for (m = 0 ; m < delimiter[level]->ndelimiters ; m++)
        for (n = 0 ; n < delimiter[level]->nscope[m] ; n++)
          if (delimiter[level]->scope[m][n].restart)
            delimiter[level]->scope_seq[m] = 0;
      level++;
    }
  }
}


